# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os.path as osp
import tempfile

import mmcv
import pytest

from otx.algorithms.detection.utils.data import (
    CocoDataset,
    LoadAnnotations,
    find_label_by_name,
    format_list_to_str,
    get_anchor_boxes,
    get_classes_from_annotation,
    get_sizes_from_dataset_entity,
    load_dataset_items_coco_format,
)
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)

# TODO: Need to add adaptive_tile_params unit-test


def _create_dummy_coco_json(json_name):
    image = {
        "id": 0,
        "width": 640,
        "height": 640,
        "file_name": "fake_name.jpg",
    }

    annotation_1 = {
        "id": 1,
        "image_id": 0,
        "category_id": 0,
        "area": 400,
        "bbox": [50, 60, 20, 20],
        "iscrowd": 0,
    }

    annotation_2 = {
        "id": 2,
        "image_id": 0,
        "category_id": 0,
        "area": 900,
        "bbox": [100, 120, 30, 30],
        "iscrowd": 0,
    }

    categories = [
        {
            "id": 0,
            "name": "car",
            "supercategory": "car",
        }
    ]

    fake_json = {
        "images": [image],
        "annotations": [annotation_1, annotation_2],
        "categories": categories,
    }

    mmcv.dump(fake_json, json_name)


class TestDataUtilsFunctionsInputParamsValidation:
    @e2e_pytest_unit
    def test_get_classes_from_annotation_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_classes_from_annotation" function input parameters validation

        <b>Input data:</b>
        "path" unexpected object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "get_classes_from_annotation" function
        """
        for unexpected_value in [
            # non string object is specified as "path" parameter
            1,
            # Empty string is specified as "path" parameter
            "",
            # Path to file with unexpected extension is specified as "path" parameter
            "./unexpected_extension.yaml",
            # Path to non-existing file is specified as "path" parameter
            "./non_existing.json",
            # Path with null character is specified as "path" parameter
            "./null\0char.json",
            # Path with non-printable character is specified as "path" parameter
            "./\non_printable_char.json",
        ]:
            with pytest.raises(ValueError):
                get_classes_from_annotation(path=unexpected_value)

    @e2e_pytest_unit
    def test_find_label_by_name_params_validation(self):
        """
        <b>Description:</b>
        Check "find_label_by_name" function input parameters validation

        <b>Input data:</b>
        "find_label_by_name" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "find_label_by_name" function
        """
        label = LabelEntity(name="test label", domain=Domain.DETECTION)
        correct_values_dict = {
            "labels": [label],
            "name": "test label",
            "domain": Domain.DETECTION,
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "labels" parameter
            ("labels", unexpected_int),
            # Unexpected integer is specified as nested label
            ("labels", [label, unexpected_int]),
            # Unexpected integer is specified as "name" parameter
            ("name", unexpected_int),
            # Unexpected integer is specified as "domain" parameter
            ("domain", unexpected_int),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=find_label_by_name,
        )

    @e2e_pytest_unit
    def test_load_dataset_items_coco_format_params_validation(self):
        """
        <b>Description:</b>
        Check "load_dataset_items_coco_format" function input parameters validation

        <b>Input data:</b>
        "load_dataset_items_coco_format" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "load_dataset_items_coco_format" function
        """
        tmp_dir = tempfile.TemporaryDirectory()
        fake_json_file = osp.join(tmp_dir.name, "fake_data.json")
        _create_dummy_coco_json(fake_json_file)

        label = LabelEntity(name="test label", domain=Domain.DETECTION)
        correct_values_dict = {
            "ann_file_path": fake_json_file,
            "data_root_dir": tmp_dir.name,
            "domain": Domain.DETECTION,
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "ann_file_path" parameter
            ("ann_file_path", unexpected_int),
            # Empty string is specified as "ann_file_path" parameter
            ("ann_file_path", ""),
            # Path to non-json file is specified as "ann_file_path" parameter
            ("ann_file_path", osp.join(tmp_dir.name, "non_json.jpg")),
            # Path with null character is specified as "ann_file_path" parameter
            ("ann_file_path", osp.join(tmp_dir.name, "\0fake_data.json")),
            # Path with non-printable character is specified as "ann_file_path" parameter
            ("ann_file_path", osp.join(tmp_dir.name, "\nfake_data.json")),
            # Path to non-existing file is specified as "ann_file_path" parameter
            ("ann_file_path", osp.join(tmp_dir.name, "non_existing.json")),
            # Unexpected integer is specified as "data_root_dir" parameter
            ("data_root_dir", unexpected_int),
            # Empty string is specified as "data_root_dir" parameter
            ("data_root_dir", ""),
            # Path with null character is specified as "data_root_dir" parameter
            ("data_root_dir", "./\0null_char"),
            # Path with non-printable character is specified as "data_root_dir" parameter
            ("data_root_dir", "./\non_printable_char"),
            # Unexpected integer is specified as "domain" parameter
            ("domain", unexpected_int),
            # Unexpected integer is specified as "subset" parameter
            ("subset", unexpected_int),
            # Unexpected integer is specified as "labels_list" parameter
            ("labels_list", unexpected_int),
            # Unexpected integer is specified as nested label
            ("labels_list", [label, unexpected_int]),
            # Unexpected string is specified as "with_mask" parameter
            ("with_mask", "unexpected string"),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=load_dataset_items_coco_format,
        )

    @e2e_pytest_unit
    def test_get_sizes_from_dataset_entity_params_validation(self):
        """
        <b>Description:</b>
        Check "get_sizes_from_dataset_entity" function input parameters validation

        <b>Input data:</b>
        "get_sizes_from_dataset_entity" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_sizes_from_dataset_entity" function
        """
        correct_values_dict = {
            "dataset": DatasetEntity(),
            "target_wh": [(0.1, 0.1)],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "dataset" parameter
            ("dataset", unexpected_int),
            # Unexpected integer is specified as "target_wh" parameter
            ("target_wh", unexpected_int),
            # Unexpected integer is specified as nested target_wh
            ("target_wh", [(0.1, 0.1), unexpected_int]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=get_sizes_from_dataset_entity,
        )

    @e2e_pytest_unit
    def test_format_list_to_str_params_validation(self):
        """
        <b>Description:</b>
        Check "format_list_to_str" function input parameters validation

        <b>Input data:</b>
        "value_lists" unexpected type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "format_list_to_str" function
        """
        with pytest.raises(ValueError):
            format_list_to_str(value_lists="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_get_anchor_boxes_params_validation(self):
        """
        <b>Description:</b>
        Check "get_anchor_boxes" function input parameters validation

        <b>Input data:</b>
        "get_anchor_boxes" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_anchor_boxes" function
        """
        correct_values_dict = {
            "wh_stats": [("wh_stat_1", 1), ("wh_stat_2", 2)],
            "group_as": [0, 1, 2],
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "wh_stats" parameter
            ("wh_stats", unexpected_str),
            # Unexpected string is specified as nested "wh_stat"
            ("wh_stats", [("wh_stat_1", 1), unexpected_str]),
            # Unexpected string is specified as "group_as" parameter
            ("group_as", unexpected_str),
            # Unexpected string is specified as nested "group_as"
            ("group_as", [0, 1, 2, unexpected_str]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=get_anchor_boxes,
        )


class TestLoadAnnotationsInputParamsValidation:
    @e2e_pytest_unit
    def test_load_annotations_init_params_validation(self):
        """
        <b>Description:</b>
        Check LoadAnnotations object initialization parameters validation

        <b>Input data:</b>
        LoadAnnotations object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        LoadAnnotations initialization parameter
        """
        for parameter in ["with_bbox", "with_label", "with_mask"]:
            with pytest.raises(ValueError):
                LoadAnnotations(**{parameter: "unexpected string"})

    @e2e_pytest_unit
    def test_load_annotations_call_params_validation(self):
        """
        <b>Description:</b>
        Check LoadAnnotations object "__call__" method input parameters validation

        <b>Input data:</b>
        "results" parameter with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__call__" method
        """
        load_annotations = LoadAnnotations()
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "results" parameter
            unexpected_int,
            # Unexpected integer is specified as "results" dictionary key
            {"result_1": "some results", unexpected_int: "unexpected results"},
        ]:
            with pytest.raises(ValueError):
                load_annotations(results=unexpected_value)


class TestCocoDatasetInputParamsValidation:
    @staticmethod
    def create_fake_json_file():
        tmp_dir = tempfile.TemporaryDirectory()
        fake_json_file = osp.join(tmp_dir.name, "fake_data.json")
        _create_dummy_coco_json(fake_json_file)
        return fake_json_file

    @staticmethod
    def dataset():
        tmp_dir = tempfile.TemporaryDirectory()
        fake_json_file = osp.join(tmp_dir.name, "fake_data.json")
        _create_dummy_coco_json(fake_json_file)
        return CocoDataset(fake_json_file)

    @e2e_pytest_unit
    def test_coco_dataset_init_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object initialization parameters validation

        <b>Input data:</b>
        CocoDataset object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        CocoDataset object initialization parameter
        """
        tmp_dir = tempfile.TemporaryDirectory()
        fake_json_file = osp.join(tmp_dir.name, "fake_data.json")
        _create_dummy_coco_json(fake_json_file)

        correct_values_dict = {
            "ann_file": fake_json_file,
        }
        unexpected_str = "unexpected string"
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "ann_file" parameter
            ("ann_file", unexpected_int),
            # Empty string is specified as "ann_file" parameter
            ("ann_file", ""),
            # Path to non-json file is specified as "ann_file" parameter
            ("ann_file", osp.join(tmp_dir.name, "non_json.jpg")),
            # Path with null character is specified as "ann_file" parameter
            ("ann_file", osp.join(tmp_dir.name, "\0fake_data.json")),
            # Path with non-printable character is specified as "ann_file" parameter
            ("ann_file", osp.join(tmp_dir.name, "\nfake_data.json")),
            # Path to non-existing file is specified as "ann_file" parameter
            ("ann_file", osp.join(tmp_dir.name, "non_existing.json")),
            # Unexpected integer is specified as "classes" parameter
            ("classes", unexpected_int),
            # Unexpected integer is specified nested class
            ("classes", ["class_1", unexpected_int]),
            # Unexpected integer is specified as "data_root" parameter
            ("data_root", unexpected_int),
            # Empty string is specified as "data_root" parameter
            ("data_root", ""),
            # Path with null character is specified as "data_root" parameter
            ("data_root", "./\0null_char"),
            # Path with non-printable character is specified as "data_root" parameter
            ("data_root", "./\non_printable_char"),
            # Unexpected integer is specified as "img_prefix" parameter
            ("img_prefix", unexpected_int),
            # Unexpected string is specified as "test_mode" parameter
            ("test_mode", unexpected_str),
            # Unexpected string is specified as "filter_empty_gt" parameter
            ("filter_empty_gt", unexpected_str),
            # Unexpected string is specified as "min_size" parameter
            ("min_size", unexpected_str),
            # Unexpected string is specified as "with_mask" parameter
            ("with_mask", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=CocoDataset,
        )

    @e2e_pytest_unit
    def test_coco_dataset_pre_pipeline_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "pre_pipeline" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "results" parameter with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "pre_pipeline" method
        """
        dataset = self.dataset()
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "results" parameter
            unexpected_int,
            # Unexpected integer is specified as "results" dictionary key
            {"result_1": "some results", unexpected_int: "unexpected results"},
        ]:
            with pytest.raises(ValueError):
                dataset.pre_pipeline(results=unexpected_value)

    @e2e_pytest_unit
    def test_coco_dataset_get_item_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "__getitem__" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__getitem__" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.__getitem__(idx="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_coco_dataset_prepare_img_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "prepare_img" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "prepare_img" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.prepare_img(idx="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_coco_dataset_get_classes_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "get_classes" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "classes" parameter with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_classes" method
        """
        dataset = self.dataset()
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "classes" parameter
            unexpected_int,
            # Unexpected integer is specified as nested "classes" element
            ["class_1", unexpected_int],
        ]:
            with pytest.raises(ValueError):
                dataset.get_classes(classes=unexpected_value)  # type: ignore

    @e2e_pytest_unit
    def test_coco_dataset_load_annotations_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "load_annotations" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "ann_file" unexpected object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "load_annotations" method
        """
        dataset = self.dataset()
        for unexpected_value in [
            # Unexpected integer is specified as "ann_file" parameter
            1,
            # Empty string is specified as "ann_file" parameter
            "",
            # Path to non-existing file is specified as "ann_file" parameter
            "./non_existing.json",
            # Path to non-json file is specified as "ann_file" parameter
            "./unexpected_type.jpg",
            # Path Null character is specified in "ann_file" parameter
            "./null\0char.json",
            # Path with non-printable character is specified as "input_config" parameter
            "./null\nchar.json",
        ]:
            with pytest.raises(ValueError):
                dataset.load_annotations(ann_file=unexpected_value)

    @e2e_pytest_unit
    def test_coco_dataset_get_ann_info_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "get_ann_info" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_ann_info" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.get_ann_info(idx="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_coco_dataset_get_cat_ids_params_validation(self):
        """
        <b>Description:</b>
        Check CocoDataset object "get_cat_ids" method input parameters validation

        <b>Input data:</b>
        CocoDataset object, "idx" non-integer type parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_cat_ids" method
        """
        dataset = self.dataset()
        with pytest.raises(ValueError):
            dataset.get_cat_ids(idx="unexpected string")  # type: ignore



"""Helper utils for tests."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from otx.api.configuration.helper import create as create_hyper_parameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model_template import (
    ModelTemplate,
    TaskType,
    parse_model_template,
)
from otx.api.entities.task_environment import TaskEnvironment


def get_model_template(model_name: str, task_type: str = "classification") -> ModelTemplate:
    template_file_root = Path("otx", "algorithms", "anomaly", "configs", task_type, model_name)
    template_file_path = (
        template_file_root / "template.yaml"
        if (template_file_root / "template.yaml").exists()
        else template_file_root / "template_experimental.yaml"
    )
    model_template: ModelTemplate = parse_model_template(str(template_file_path))
    return model_template


def create_task_environment(dataset: DatasetEntity, task_type: TaskType) -> TaskEnvironment:
    # get padim model template
    padim_template = get_model_template("padim")
    padim_template.task_type = task_type
    hyper_parameters = create_hyper_parameters(padim_template.hyper_parameters.data)
    labels = dataset.get_labels()
    label_schema = LabelSchemaEntity.from_labels(labels)

    return TaskEnvironment(
        model_template=padim_template, model=None, hyper_parameters=hyper_parameters, label_schema=label_schema
    )



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.segmentation.adapters.mmseg.data.dataset import MPASegDataset
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def label_entity(name="test label") -> LabelEntity:
    return LabelEntity(name=name, domain=Domain.SEGMENTATION)


def dataset_item() -> DatasetItemEntity:
    image: Image = Image(data=np.random.randint(low=0, high=255, size=(10, 16, 3)))
    annotation: Annotation = Annotation(shape=Rectangle.generate_full_box(), labels=[ScoredLabel(label_entity())])
    annotation_scene: AnnotationSceneEntity = AnnotationSceneEntity(
        annotations=[annotation], kind=AnnotationSceneKind.ANNOTATION
    )
    return DatasetItemEntity(media=image, annotation_scene=annotation_scene)


class TestMPASegDataset:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.otx_dataset: DatasetEntity = DatasetEntity(items=[dataset_item()])
        self.pipeline: list[dict] = [{"type": "LoadImageFromOTXDataset", "to_float32": True}]
        self.classes: list[str] = ["class_1", "class_2"]

        self.dataset: MPASegDataset = MPASegDataset(
            otx_dataset=self.otx_dataset,
            pipeline=self.pipeline,
            labels=[label_entity(name) for name in self.classes],
            new_classes=self.classes,
        )

    @e2e_pytest_unit
    def test_mpasegdataset_initialization(self) -> None:
        assert self.dataset.otx_dataset == self.otx_dataset
        assert self.dataset.CLASSES == ["background"] + self.classes

        # Check if img_indices are generated properly
        assert hasattr(self.dataset, "img_indices")

        # Check if label_map is created as expected
        assert self.dataset.label_map == {0: 0, 1: 1, 2: 2}

    @e2e_pytest_unit
    def test_getitem_method(self) -> None:
        data_item: dict = self.dataset[0]
        assert "dataset_item" in data_item
        assert "ann_info" in data_item

    @e2e_pytest_unit
    def test_getting_annotation_info(self) -> None:
        annotation_info: dict = self.dataset.get_ann_info(0)
        assert "gt_semantic_seg" in annotation_info

    @e2e_pytest_unit
    def test_get_gt_seg_maps(self) -> None:
        gt_seg_map: np.ndarray = self.dataset.get_gt_seg_maps()[0]
        assert np.equal(gt_seg_map, np.zeros((10, 16))).all()



"""Multi-layer Perceptron (MLP) for Self-supervised learning methods.

This MLP consists of fc (conv) - norm - relu - fc (conv).
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=missing-module-docstring, dangerous-default-value
from typing import Any, Dict, List, Tuple, Union

import torch
from mmcls.models.builder import NECKS
from mmcv.cnn import build_norm_layer, kaiming_init, normal_init
from torch import nn


@NECKS.register_module()
class SelfSLMLP(nn.Module):
    """The SelfSLMLP neck: fc/conv-bn-relu-fc/conv.

    Args:
        in_channels (int): The number of feature output channels from backbone.
        hid_channels (int): The number of channels for a hidden layer.
        out_channels (int): The number of output channels of SelfSLMLP.
        norm_cfg (dict): Normalize configuration. Default: dict(type="BN1d").
        use_conv (bool): Whether using conv instead of fc. Default: False.
        with_avg_pool (bool): Whether using average pooling before passing MLP.
                              Default: True.
    """

    def __init__(
        self,
        in_channels: int,
        hid_channels: int,
        out_channels: int,
        norm_cfg: Dict[str, Any] = dict(type="BN1d"),
        use_conv: bool = False,
        with_avg_pool: bool = True,
    ):

        super().__init__()

        self.with_avg_pool = with_avg_pool
        if with_avg_pool:
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.use_conv = use_conv
        if use_conv:
            self.mlp = nn.Sequential(
                nn.Conv2d(in_channels, hid_channels, 1),
                build_norm_layer(norm_cfg, hid_channels)[1],
                nn.ReLU(inplace=True),
                nn.Conv2d(hid_channels, out_channels, 1),
            )
        else:
            self.mlp = nn.Sequential(
                nn.Linear(in_channels, hid_channels),
                build_norm_layer(norm_cfg, hid_channels)[1],
                nn.ReLU(inplace=True),
                nn.Linear(hid_channels, out_channels),
            )

    def init_weights(self, init_linear: str = "normal", std: float = 0.01, bias: float = 0.0):
        """Initialize SelfSLMLP weights.

        Args:
            init_linear (str): Option to initialize weights. Default: "normal".
            std (float): Standard deviation for normal initialization. Default: 0.01.
            bias (float): Bias for normal initialization. Default: 0.
        """
        if init_linear not in ["normal", "kaiming"]:
            raise ValueError(f"Undefined init_linear: {init_linear}")
        for m in self.modules():  # pylint: disable=invalid-name
            if isinstance(m, nn.Linear):
                if init_linear == "normal":
                    normal_init(m, std=std, bias=bias)
                else:
                    kaiming_init(m, mode="fan_in", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.GroupNorm, nn.SyncBatchNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x: Union[torch.Tensor, Tuple, List]):
        """Forward SelfSLMLP.

        Args:
            x (Tensor, tuple, list): Inputs to pass MLP.
                If a type of the inputs is tuple or list, just use the last index.

        Return:
            Tensor: Features passed SelfSLMLP.
        """
        if isinstance(x, (tuple, list)):
            # using last output
            x = x[-1]
        if not isinstance(x, torch.Tensor):
            raise TypeError("neck inputs should be tuple or torch.tensor")
        if self.with_avg_pool:
            x = self.avgpool(x)
        if self.use_conv:  # pylint: disable=no-else-return
            return self.mlp(x)
        else:
            return self.mlp(x.view(x.size(0), -1))



import mmcv

from otx.mpa.builder import build
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_build_with_stages(mocker):
    cfg = mmcv.ConfigDict(
        stages=[mocker.MagicMock()],
        type=mocker.MagicMock(),
        workflow_hooks=[mocker.MagicMock()],
    )
    mocker.patch("otx.mpa.builder.build_workflow_hook")
    mock_build_from_cfg = mocker.patch("otx.mpa.builder.build_from_cfg")
    mock_workflow = mocker.patch("otx.mpa.builder.Workflow")
    mocker.patch("otx.mpa.builder.config_logger")
    mocker.patch("os.makedirs")
    mocker.patch("os.unlink")
    mocker.patch("os.symlink")

    build(cfg)

    mock_build_from_cfg.assert_called()
    mock_workflow.assert_called_once()


@e2e_pytest_unit
def test_build_without_stages(mocker):
    cfg = mmcv.ConfigDict()

    mocker.patch("otx.mpa.builder.get_available_types", return_value="MockStage")
    mock_build_from_cfg = mocker.patch("otx.mpa.builder.build_from_cfg")

    build(cfg, None, "MockStage")

    mock_build_from_cfg.assert_called_once()



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import otx.algorithms.classification.adapters.mmcls.nncf.registers  # noqa: F401
from otx.algorithms.common.adapters.nncf.utils import is_nncf_enabled
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_registers():
    if is_nncf_enabled():
        from nncf.torch.layers import UNWRAPPED_USER_MODULES
        from timm.models.layers.conv2d_same import Conv2dSame

        assert Conv2dSame in UNWRAPPED_USER_MODULES.registry_dict.values()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest
from openvino.model_zoo.model_api.models.classification import Classification

from otx.algorithms.classification.adapters.openvino.model_wrappers import (
    OTXClassification,
)
from otx.algorithms.classification.utils.labels_utils import get_multihead_class_info
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.classification.test_helper import (
    generate_cls_dataset,
    generate_label_schema,
)


class MockClassification(OTXClassification, Classification):
    def __init__(self):
        self.out_layer_name = "logits"
        hierarchical_dataset = generate_cls_dataset(hierarchical=True)
        label_schema = generate_label_schema(hierarchical_dataset.get_labels(), multilabel=False, hierarchical=True)
        self.multihead_class_info = get_multihead_class_info(label_schema)


class TestOTXClassification:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.cls_ov_model = MockClassification()
        self.fake_logits = np.array([-0.12, 0.92])
        self.fake_logits_hierarchical = np.array([3.01, -2.75, -0.63, 0.74, 0.017, -0.12, -0.42, -0.88])

    @pytest.mark.parametrize("multilabel, hierarchical", [(False, False), (True, False), (False, True)])
    @e2e_pytest_unit
    def test_postprocess(self, multilabel, hierarchical):
        self.cls_ov_model.multilabel = multilabel
        self.cls_ov_model.hierarchical = hierarchical
        fake_model_out = {"logits": self.fake_logits}
        if self.cls_ov_model.hierarchical:
            fake_model_out = {"logits": self.fake_logits_hierarchical}

        fake_metadata = {"original_shape": (254, 320, 3), "resized_shape": (224, 224, 3)}
        prediction = self.cls_ov_model.postprocess(fake_model_out, fake_metadata)
        if self.cls_ov_model.hierarchical:
            assert len(prediction) == 3
        else:
            assert prediction[0][0] == 1
            assert len(prediction) == 1

    @pytest.mark.parametrize("multilabel, hierarchical", [(False, False), (True, False), (False, True)])
    @e2e_pytest_unit
    def test_postprocess_aux_outputs(self, multilabel, hierarchical):
        self.cls_ov_model.multilabel = multilabel
        self.cls_ov_model.hierarchical = hierarchical
        fake_map = np.random.rand(1, 100, 100)
        fake_feature_vector = np.random.rand(1, 100)
        fake_logits = np.array([-0.012, 0.092])
        if self.cls_ov_model.hierarchical:
            fake_logits = np.array(
                [3.018761, -2.7551947, -0.6320721, 0.7461215, 0.01764688, -0.1284887, -0.42783177, -0.8848756]
            )

        fake_output = {"logits": fake_logits, "feature_vector": fake_feature_vector, "saliency_map": fake_map}
        fake_metadata = {"original_shape": (254, 320, 3), "resized_shape": (224, 224, 3)}
        out = self.cls_ov_model.postprocess_aux_outputs(fake_output, fake_metadata)
        assert len(out) == 4
        for item in out:
            assert item is not None



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy

import torch
import torch.nn as nn
from mmcv.runner import HOOKS, Hook

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class ModelEmaV2Hook(Hook):
    """
    Source model paramters would be exponentially averaged
    onto destination model pararmeters on given intervals
        .. math::
            \text{Xema_{t+1}} = (1 - \text{momentum}) \times
            \text{Xema_{t}} +  \text{momentum} \times X_t
    Args:
        ema_decay (float): EMA decay value used for updating ema parameter.
            Defaults to 0.999.
        interval (int): Update ema parameter every interval iteration.
            Defaults to 1.
        start_epoch (int): During initial a few epochs, we just copy values
            to update ema parameters. Defaults to 5.
        dataset_len_thr (int): number of train images in the dataset when to enable the EMA hook
    """

    def __init__(self, ema_decay=0.9995, interval=1, start_epoch=0, dataset_len_thr=2000, **kwargs):
        super().__init__(**kwargs)
        self.ema_decay = ema_decay
        self.interval = interval
        self.start_epoch = start_epoch
        self.dataset_len_thr = dataset_len_thr

    def before_train_epoch(self, runner):
        if not hasattr(self, "use_ema"):
            self.use_ema = len(runner.data_loader.dataset) > self.dataset_len_thr

        if self.use_ema and not hasattr(runner, "ema_model"):
            model = runner.model
            ema_model = ModelEmaV2(model, decay=self.ema_decay, dataset_len_thr=self.dataset_len_thr)
            runner.ema_model = ema_model

    def before_run(self, runner):
        logger.info("\t* EMA V2 Enable")

    def after_train_iter(self, runner):
        """Update ema parameter every self.interval iterations."""
        if not self.use_ema:
            return

        if runner.iter % self.interval != 0:
            # Skip update
            return

        if runner.epoch < self.start_epoch:
            # Just copy parameters before start epoch
            return

        runner.ema_model.update()


class ModelEmaV2(nn.Module):
    """Model Exponential Moving Average V2
    Keep a moving average of everything in the model state_dict (parameters and buffers).
    V2 of this module is simpler, it does not match params/buffers based on name but simply
    iterates in order. It works with torchscript (JIT of full model).
    This is intended to allow functionality like
    https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    A smoothed version of the weights is necessary for some training schemes to perform well.
    E.g. Google's hyper-params for training MNASNet, MobileNet-V3, EfficientNet, etc that use
    RMSprop with a short 2.4-3 epoch decay period and slow LR decay rate of .96-.99 requires EMA
    smoothing of weights to match results. Pay attention to the decay constant you are using
    relative to your update count per epoch.
    To keep EMA from using GPU resources, set device='cpu'. This will save a bit of memory but
    disable validation of the EMA weights. Validation will have to be done manually in a separate
    process, or after the training stops converging.
    This class is sensitive where it is initialized in the sequence of model init,
    GPU assignment and distributed training wrappers.
    """

    def __init__(self, model, decay=0.9999, dataset_len_thr=None, device=None):
        super(ModelEmaV2, self).__init__()
        # make a copy of the model for accumulating moving average of weights
        self.module = deepcopy(model)
        self.module.eval()
        self.src_model = model.state_dict()
        self.dst_model = self.module.state_dict()
        self.decay = decay
        self.device = device  # perform ema on different device from model if set
        self.dataset_len_thr = dataset_len_thr
        if self.device is not None:
            self.module.to(device=device)

    def _update(self, update_fn):
        with torch.no_grad():
            for ema_v, model_v in zip(self.dst_model.values(), self.src_model.values()):
                if self.device is not None:
                    model_v = model_v.to(device=self.device)
                ema_v.copy_(update_fn(ema_v, model_v))

    def update(self):
        self._update(update_fn=lambda e, m: self.decay * e + (1.0 - self.decay) * m)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
import torch.nn.functional as F
from mmcls.models.builder import LOSSES
from mmcls.models.losses.utils import weight_reduce_loss


def cross_entropy(pred, label, weight=None, reduction="mean", avg_factor=None, class_weight=None, ignore_index=None):
    # element-wise losses
    if ignore_index is not None:
        loss = F.cross_entropy(pred, label, reduction="none", weight=class_weight, ignore_index=ignore_index)
    else:
        loss = F.cross_entropy(pred, label, reduction="none", weight=class_weight)

    # apply weights and do the reduction
    if weight is not None:
        weight = weight.float()
    loss = weight_reduce_loss(loss, weight=weight, reduction=reduction, avg_factor=avg_factor)

    return loss


@LOSSES.register_module()
class CrossEntropyLossWithIgnore(nn.Module):
    def __init__(self, reduction="mean", loss_weight=1.0, ignore_index=None):
        super(CrossEntropyLossWithIgnore, self).__init__()
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.ignore_index = ignore_index

        self.cls_criterion = cross_entropy

    def forward(self, cls_score, label, weight=None, avg_factor=None, reduction_override=None, **kwargs):
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = reduction_override if reduction_override else self.reduction
        loss_cls = self.loss_weight * self.cls_criterion(
            cls_score,
            label,
            weight,
            ignore_index=self.ignore_index,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs
        )
        return loss_cls



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from copy import deepcopy
from typing import Any, List, Union

import cv2
import numpy as np
import pytest
from PIL import Image

from otx.mpa.modules.datasets.pipelines.transforms.augments import (
    Augments,
    CythonAugments,
)


@pytest.fixture
def images() -> List[Image.Image]:
    n_seed = 3003
    n_imgs = 4
    n_shapes = 4
    img_size = 50
    size = [img_size, img_size, 3]

    np.random.seed(n_seed)

    imgs = []
    for _ in range(n_imgs):
        img = np.full(size, 0, dtype=np.uint8)
        for _ in range(n_shapes):
            position = np.random.randint(0, 50, size=[2]).tolist()
            color = np.random.randint(0, 256, size=[3]).tolist()
            marker_type = np.random.randint(0, 7)
            img = cv2.drawMarker(img, position, color, marker_type, thickness=5)
        imgs += [Image.fromarray(img)]

    return imgs


EXACT_EQUAL_TESTS = [
    ("autocontrast", []),
    ("equalize", []),
    ("solarize", [64, 128, 196]),
    ("posterize", [1, 4, 7]),
]


@pytest.mark.parametrize("func,params", EXACT_EQUAL_TESTS)
def test_exact_equal(images: List[Image.Image], func: str, params: List[Any]):
    for img in images:
        for param in params:
            grt = getattr(Augments, func)(deepcopy(img), param)
            tst = getattr(CythonAugments, func)(deepcopy(img), param)

            assert np.array_equal(np.asarray(grt), np.asarray(tst))


APPROX_EQUAL_TESTS = [
    ("color", [0.1, 0.5, 0.9], 1),
    ("contrast", [0.1, 0.5, 0.9], 1),
    ("brightness", [0.1, 0.5, 0.9], 1),
    ("sharpness", [0.25, 0.75, 1.25, 1.75], 1),
    ("rotate", [-35, -15, 15, 35], 1),
    ("shear_x", [-0.8, -0.3, -0.3, 0.8], 1),
    ("shear_y", [-0.8, -0.3, -0.3, 0.8], 1),
    ("translate_x_rel", [-0.8, -0.3, -0.3, 0.8], 1),
    ("translate_y_rel", [-0.8, -0.3, -0.3, 0.8], 1),
]


@pytest.mark.parametrize("func,params,tol", APPROX_EQUAL_TESTS)
def test_approx_equal(images: List[Image.Image], func: str, params: List[Any], tol: Union[float, int]):
    for img in images:
        for param in params:
            grt = getattr(Augments, func)(deepcopy(img), param)
            tst = getattr(CythonAugments, func)(deepcopy(img), param)
            grt = np.array(grt).astype(np.float32)
            tst = np.array(tst).astype(np.float32)
            med = np.median(grt - tst)
            assert med <= tol



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
from mmcls.models.builder import LOSSES
from mmcls.models.losses.utils import weight_reduce_loss


def asymmetric_angular_loss_with_ignore(
    pred,
    target,
    valid_label_mask=None,
    weight=None,
    gamma_pos=0.0,
    gamma_neg=1.0,
    clip=0.05,
    k=0.8,
    reduction="mean",
    avg_factor=None,
):
    """asymmetric angular loss
    Args:
        pred (torch.Tensor): The prediction with shape (N, *).
        target (torch.Tensor): The ground truth label of the prediction with
            shape (N, *).
        weight (torch.Tensor, optional): Sample-wise loss weight with shape
            (N, ). Dafaults to None.
        gamma_pos (float): positive focusing parameter. Defaults to 0.0.
        gamma_neg (float): Negative focusing parameter. We usually set
            gamma_neg > gamma_pos. Defaults to 1.0.
        k (float): positive balance parameter. Defaults to 0.8.
        clip (float, optional): Probability margin. Defaults to 0.05.
        reduction (str): The method used to reduce the loss.
            Options are "none", "mean" and "sum". If reduction is 'none' , loss
             is same shape as pred and label. Defaults to 'mean'.
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
    Returns:
        torch.Tensor: Loss.
    """
    assert pred.shape == target.shape, "pred and target should be in the same shape."

    eps = 1e-8
    target = target.type_as(pred)
    anti_target = 1 - target

    xs_pos = torch.sigmoid(pred)
    xs_neg = torch.sigmoid(-pred)

    if clip > 0:
        xs_neg = (xs_neg + clip).clamp(max=1)

    asymmetric_focus = gamma_pos > 0 or gamma_neg > 0
    if asymmetric_focus:
        pt0 = xs_neg * target
        pt1 = xs_pos * anti_target
        pt = pt0 + pt1
        one_sided_gamma = gamma_pos * target + gamma_neg * anti_target
        one_sided_w = torch.pow(pt, one_sided_gamma)

    loss = -k * target * torch.log(xs_pos.clamp(min=eps)) - (1 - k) * anti_target * torch.log(xs_neg.clamp(min=eps))

    if asymmetric_focus:
        loss *= one_sided_w

    if valid_label_mask is not None:
        loss = loss * valid_label_mask

    if weight is not None:
        assert weight.dim() == 1
        weight = weight.float()
        if pred.dim() > 1:
            weight = weight.reshape(-1, 1)
    if reduction != "mean":
        avg_factor = None
    loss = weight_reduce_loss(loss, weight, reduction, avg_factor)
    return loss


@LOSSES.register_module()
class AsymmetricAngularLossWithIgnore(nn.Module):
    """Asymmetric angular loss
    Args:
        gamma_pos (float): positive focusing parameter.
            Defaults to 0.0.
        gamma_neg (float): Negative focusing parameter. We
            usually set gamma_neg > gamma_pos. Defaults to 1.0.
        k (float): positive balance parameter. Defaults to 0.8.
        clip (float): Probability margin. Defaults to 0.05.
        reduction (str): The method used to reduce the loss into
            a scalar.
        loss_weight (float): Weight of loss. Defaults to 1.0.
    """

    def __init__(self, gamma_pos=0.0, gamma_neg=1.0, k=0.8, clip=0.05, reduction="mean", loss_weight=1.0):
        super().__init__()
        self.gamma_pos = gamma_pos
        self.gamma_neg = gamma_neg
        self.k = k
        self.clip = clip
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self, pred, target, valid_label_mask=None, weight=None, avg_factor=None, reduction_override=None):
        """asymmetric angular loss"""
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = reduction_override if reduction_override else self.reduction
        loss_cls = self.loss_weight * asymmetric_angular_loss_with_ignore(
            pred,
            target,
            valid_label_mask,
            weight,
            gamma_pos=self.gamma_pos,
            gamma_neg=self.gamma_neg,
            k=self.k,
            clip=self.clip,
            reduction=reduction,
            avg_factor=avg_factor,
        )
        return loss_cls



"""OTX Adapters - mmseg."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .data import MPASegDataset
from .models import DetConB, DetConLoss, SelfSLMLP, SupConDetConB

# fmt: off
# isort: off
# FIXME: openvino pot library adds stream handlers to root logger
# which makes annoying duplicated logging
# pylint: disable=no-name-in-module,wrong-import-order
from mmseg.utils import get_root_logger  # type: ignore # (false positive)
get_root_logger().propagate = False
# fmt: off
# isort: on

__all__ = ["MPASegDataset", "DetConLoss", "SelfSLMLP", "DetConB", "SupConDetConB"]



"""Collection of utils for task implementation in Classification Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import math
from typing import List, Optional, Union

from mmcv import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_keys,
    get_configs_by_pairs,
    get_dataset_configs,
    get_meta_keys,
    is_epoch_based_runner,
    patch_color_conversion,
    prepare_work_dir,
    remove_from_config,
    remove_from_configs_by_type,
    update_config,
)
from otx.api.entities.label import Domain, LabelEntity
from otx.api.utils.argument_checks import (
    DirectoryPathCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

logger = get_logger()


@check_input_parameters_type({"work_dir": DirectoryPathCheck})
def patch_config(
    config: Config,
    work_dir: str,
    labels: List[LabelEntity],
):  # pylint: disable=too-many-branches
    """Update config function."""

    # Add training cancelation hook.
    if "custom_hooks" not in config:
        config.custom_hooks = []
    if "CancelTrainingHook" not in {hook.type for hook in config.custom_hooks}:
        config.custom_hooks.append(ConfigDict({"type": "CancelTrainingHook"}))

    # Remove high level data pipelines definition leaving them only inside `data` section.
    remove_from_config(config, "train_pipeline")
    remove_from_config(config, "test_pipeline")
    remove_from_config(config, "train_pipeline_strong")
    # Remove cancel interface hook
    remove_from_configs_by_type(config.custom_hooks, "CancelInterfaceHook")

    config.checkpoint_config.max_keep_ckpts = 5
    config.checkpoint_config.interval = config.evaluation.get("interval", 1)

    set_data_classes(config, labels)

    config.gpu_ids = range(1)
    config.work_dir = work_dir


@check_input_parameters_type()
def patch_model_config(
    config: Config,
    labels: List[LabelEntity],
):
    """Patch model config."""
    set_num_classes(config, len(labels))


@check_input_parameters_type()
def patch_adaptive_repeat_dataset(
    config: Union[Config, ConfigDict],
    num_samples: int,
    decay: float = -0.002,
    factor: float = 30,
):
    """Patch the repeat times and training epochs adatively.

    Frequent dataloading inits and evaluation slow down training when the
    sample size is small. Adjusting epoch and dataset repetition based on
    empirical exponential decay improves the training time by applying high
    repeat value to small sample size dataset and low repeat value to large
    sample.

    :param config: mmcv config
    :param num_samples: number of training samples
    :param decay: decaying rate
    :param factor: base repeat factor
    """
    data_train = config.data.train
    if data_train.type == "RepeatDataset" and getattr(data_train, "adaptive_repeat_times", False):
        if is_epoch_based_runner(config.runner):
            cur_epoch = config.runner.max_epochs
            new_repeat = max(round(math.exp(decay * num_samples) * factor), 1)
            new_epoch = math.ceil(cur_epoch / new_repeat)
            if new_epoch == 1:
                return
            config.runner.max_epochs = new_epoch
            data_train.times = new_repeat


@check_input_parameters_type()
def prepare_for_training(
    config: Union[Config, ConfigDict],
    data_config: ConfigDict,
) -> Union[Config, ConfigDict]:
    """Prepare configs for training phase."""
    prepare_work_dir(config)

    train_num_samples = 0
    for subset in ["train", "val", "test"]:
        data_config_ = data_config.data.get(subset)
        config_ = config.data.get(subset)
        if data_config_ is None:
            continue
        for key in ["otx_dataset", "labels"]:
            found = get_configs_by_keys(data_config_, key, return_path=True)
            if len(found) == 0:
                continue
            assert len(found) == 1
            if subset == "train" and key == "otx_dataset":
                found_value = list(found.values())[0]
                if found_value:
                    train_num_samples = len(found_value)
            update_config(config_, found)

    if train_num_samples > 0:
        patch_adaptive_repeat_dataset(config, train_num_samples)

    return config


@check_input_parameters_type()
def set_data_classes(config: Config, labels: List[LabelEntity]):
    """Setter data classes into config."""
    # Save labels in data configs.
    for subset in ("train", "val", "test"):
        for cfg in get_dataset_configs(config, subset):
            cfg.labels = labels


@check_input_parameters_type()
def set_num_classes(config: Config, num_classes: int):
    """Set num classes."""
    head_names = ["head"]
    for head_name in head_names:
        if head_name in config.model:
            config.model[head_name].num_classes = num_classes


@check_input_parameters_type()
def patch_datasets(
    config: Config,
    domain: Domain = Domain.CLASSIFICATION,
    subsets: Optional[List[str]] = None,
    **kwargs,
):
    """Update dataset configs."""
    assert "data" in config
    assert "type" in kwargs

    if subsets is None:
        subsets = ["train", "val", "test", "unlabeled"]

    def update_pipeline(cfg):
        if subset == "train":
            for collect_cfg in get_configs_by_pairs(cfg, dict(type="Collect")):
                get_meta_keys(collect_cfg)

    for subset in subsets:
        if subset not in config.data:
            continue
        config.data[f"{subset}_dataloader"] = config.data.get(f"{subset}_dataloader", ConfigDict())

        # For stable hierarchical information indexing
        if subset == "train" and kwargs["type"] == "OTXHierarchicalClsDataset":
            config.data[f"{subset}_dataloader"].drop_last = True

        cfgs = get_dataset_configs(config, subset)
        for cfg in cfgs:
            cfg.domain = domain
            cfg.otx_dataset = None
            cfg.labels = None
            cfg.update(kwargs)

            update_pipeline(cfg)

    patch_color_conversion(config)


def patch_evaluation(config: Config, task: str):
    """Patch evaluation."""
    cfg = config.get("evaluation", None)
    if cfg:
        if task == "multilabel":
            cfg.metric = ["accuracy-mlc", "mAP", "CP", "OP", "CR", "OR", "CF1", "OF1"]
            config.early_stop_metric = "mAP"
        elif task == "hierarchical":
            cfg.metric = ["MHAcc", "avgClsAcc", "mAP"]
            config.early_stop_metric = "MHAcc"
        elif task == "normal":
            cfg.metric = ["accuracy", "class_accuracy"]
            config.early_stop_metric = "accuracy"
        else:
            raise NotImplementedError



"""Entities containing shapes."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# from .rectangle import *
# from .circle import *
# from .polygon import *
# from .shape import *



dataset_type = "CocoDataset"
img_size = (1024, 1024)

img_norm_cfg = dict(mean=(103.53, 116.28, 123.675), std=(1.0, 1.0, 1.0), to_rgb=False)

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True, with_mask=True, poly2mask=False),
    dict(type="Resize", img_scale=img_size, keep_ratio=False),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="Pad", size_divisor=32),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels", "gt_masks"]),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=img_size,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **img_norm_cfg),
            dict(type="Pad", size_divisor=32),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]



"""OTXMaskRCNNModel & OTXSSDModel of OTX Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Dict

import numpy as np

try:
    from openvino.model_zoo.model_api.models.instance_segmentation import MaskRCNNModel
    from openvino.model_zoo.model_api.models.ssd import SSD, find_layer_by_name
    from openvino.model_zoo.model_api.models.utils import Detection
except ImportError as e:
    import warnings

    warnings.warn(f"{e}: ModelAPI was not found.")


class OTXMaskRCNNModel(MaskRCNNModel):
    """OpenVINO model wrapper for OTX MaskRCNN model."""

    __model__ = "OTX_MaskRCNN"

    def _get_outputs(self):
        output_match_dict = {}
        output_names = ["boxes", "labels", "masks", "feature_vector", "saliency_map"]
        for output_name in output_names:
            for node_name, node_meta in self.outputs.items():
                if output_name in node_meta.names:
                    output_match_dict[output_name] = node_name
                    break
        return output_match_dict

    def postprocess(self, outputs, meta):
        """Post process function for OTX MaskRCNN model."""

        # pylint: disable-msg=too-many-locals
        resize_mask = meta.get("resize_mask", True)

        # FIXME: here, batch dim of IR must be 1
        boxes = outputs[self.output_blob_name["boxes"]]
        boxes = boxes.squeeze(0)
        assert boxes.ndim == 2
        masks = outputs[self.output_blob_name["masks"]]
        masks = masks.squeeze(0)
        assert masks.ndim == 3
        classes = outputs[self.output_blob_name["labels"]].astype(np.uint32)
        classes = classes.squeeze(0)
        assert classes.ndim == 1
        if self.is_segmentoly:
            scores = outputs[self.output_blob_name["scores"]]
        else:
            scores = boxes[:, 4]
            boxes = boxes[:, :4]
            classes += 1

        # Filter out detections with low confidence.
        detections_filter = scores > self.confidence_threshold  # pylint: disable=no-member
        scores = scores[detections_filter]
        boxes = boxes[detections_filter]
        masks = masks[detections_filter]
        classes = classes[detections_filter]

        scale_x = meta["resized_shape"][1] / meta["original_shape"][1]
        scale_y = meta["resized_shape"][0] / meta["original_shape"][0]
        boxes[:, 0::2] /= scale_x
        boxes[:, 1::2] /= scale_y

        resized_masks = []
        for box, cls, raw_mask in zip(boxes, classes, masks):
            raw_cls_mask = raw_mask[cls, ...] if self.is_segmentoly else raw_mask
            if resize_mask:
                resized_masks.append(self._segm_postprocess(box, raw_cls_mask, *meta["original_shape"][:-1]))
            else:
                resized_masks.append(raw_cls_mask)

        return scores, classes, boxes, resized_masks

    def segm_postprocess(self, *args, **kwargs):
        """Post-process for segmentation masks."""
        return self._segm_postprocess(*args, **kwargs)


class OTXSSDModel(SSD):
    """OpenVINO model wrapper for OTX SSD model."""

    __model__ = "OTX_SSD"

    def __init__(self, model_adapter, configuration=None, preload=False):
        # pylint: disable-next=bad-super-call
        super(SSD, self).__init__(model_adapter, configuration, preload)
        self.image_info_blob_name = self.image_info_blob_names[0] if len(self.image_info_blob_names) == 1 else None
        self.output_parser = BatchBoxesLabelsParser(
            self.outputs,
            self.inputs[self.image_blob_name].shape[2:][::-1],
        )

    def _get_outputs(self) -> Dict:
        """Match the output names with graph node index."""
        output_match_dict = {}
        output_names = ["boxes", "labels", "feature_vector", "saliency_map"]
        for output_name in output_names:
            for node_name, node_meta in self.outputs.items():
                if output_name in node_meta.names:
                    output_match_dict[output_name] = node_name
                    break
        return output_match_dict


class BatchBoxesLabelsParser:
    """Batched output parser."""

    def __init__(self, layers, input_size, labels_layer="labels", default_label=0):
        try:
            self.labels_layer = find_layer_by_name(labels_layer, layers)
        except ValueError:
            self.labels_layer = None
            self.default_label = default_label

        self.bboxes_layer = self.find_layer_bboxes_output(layers)
        self.input_size = input_size

    @staticmethod
    def find_layer_bboxes_output(layers):
        """find_layer_bboxes_output."""
        filter_outputs = [name for name, data in layers.items() if len(data.shape) == 3 and data.shape[-1] == 5]
        if not filter_outputs:
            raise ValueError("Suitable output with bounding boxes is not found")
        if len(filter_outputs) > 1:
            raise ValueError("More than 1 candidate for output with bounding boxes.")
        return filter_outputs[0]

    def __call__(self, outputs):
        """Parse bboxes."""
        # FIXME: here, batch dim of IR must be 1
        bboxes = outputs[self.bboxes_layer]
        bboxes = bboxes.squeeze(0)
        assert bboxes.ndim == 2
        scores = bboxes[:, 4]
        bboxes = bboxes[:, :4]
        bboxes[:, 0::2] /= self.input_size[0]
        bboxes[:, 1::2] /= self.input_size[1]
        if self.labels_layer:
            labels = outputs[self.labels_layer]
        else:
            labels = np.full(len(bboxes), self.default_label, dtype=bboxes.dtype)
        labels = labels.squeeze(0)

        detections = [Detection(*bbox, score, label) for label, score, bbox in zip(labels, scores, bboxes)]
        return detections



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
from mmcv.runner import force_fp32
from mmseg.core import add_prefix
from mmseg.models.losses import accuracy
from mmseg.ops import resize

from otx.mpa.modules.utils.seg_utils import get_valid_label_mask_per_batch

from ..losses.utils import LossEqualizer


class PixelWeightsMixin(nn.Module):
    def __init__(
        self,
        enable_loss_equalizer=False,
        loss_target="gt_semantic_seg",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.enable_loss_equalizer = enable_loss_equalizer
        self.loss_target = loss_target

        self.loss_equalizer = None
        if enable_loss_equalizer:
            self.loss_equalizer = LossEqualizer()

        self.forward_output = None

    @property
    def loss_target_name(self):
        return self.loss_target

    @property
    def last_scale(self):
        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        num_losses = len(losses_decode)
        if num_losses <= 0:
            return 1.0

        loss_module = losses_decode[0]
        if not hasattr(loss_module, "last_scale"):
            return 1.0

        return loss_module.last_scale

    def set_step_params(self, init_iter, epoch_size):
        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        for loss_module in losses_decode:
            if hasattr(loss_module, "set_step_params"):
                loss_module.set_step_params(init_iter, epoch_size)

    def forward_train(
        self,
        inputs,
        img_metas,
        gt_semantic_seg,
        train_cfg,
        pixel_weights=None,
        return_logits=False,
    ):
        """Forward function for training.
        Args:
            inputs (list[Tensor]): List of multi-level img features.
            img_metas (list[dict]): List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:Collect`.
            gt_semantic_seg (Tensor): Semantic segmentation masks
                used if the architecture supports semantic segmentation task.
            train_cfg (dict): The training config.
            pixel_weights (Tensor): Pixels weights.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        seg_logits = self.forward(inputs)
        losses = self.losses(seg_logits, gt_semantic_seg, train_cfg, pixel_weights)

        if return_logits:
            logits = self.forward_output if self.forward_output is not None else seg_logits
            return losses, logits
        return losses

    @force_fp32(apply_to=("seg_logit",))
    def losses(self, seg_logit, seg_label, train_cfg, pixel_weights=None):
        """Compute segmentation loss."""

        loss = dict()

        seg_logit = resize(
            input=seg_logit,
            size=seg_label.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        seg_label = seg_label.squeeze(1)

        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        out_losses = dict()
        for loss_idx, loss_module in enumerate(losses_decode):
            loss_value, loss_meta = loss_module(seg_logit, seg_label, pixel_weights=pixel_weights)

            loss_name = loss_module.name + f"-{loss_idx}"
            out_losses[loss_name] = loss_value
            loss.update(add_prefix(loss_meta, loss_name))

        if self.enable_loss_equalizer and len(losses_decode) > 1:
            out_losses = self.loss_equalizer.reweight(out_losses)

        for loss_name, loss_value in out_losses.items():
            loss[loss_name] = loss_value

        loss["loss_seg"] = sum(out_losses.values())
        loss["acc_seg"] = accuracy(seg_logit, seg_label)

        return loss


class PixelWeightsMixin2(PixelWeightsMixin):
    def forward_train(
        self,
        inputs,
        img_metas,
        gt_semantic_seg,
        train_cfg,
        pixel_weights=None,
        return_logits=False,
    ):
        """Forward function for training.
        Args:
            inputs (list[Tensor]): List of multi-level img features.
            img_metas (list[dict]): List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', 'img_norm_cfg',
                and 'ignored_labels'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:Collect`.
            gt_semantic_seg (Tensor): Semantic segmentation masks
                used if the architecture supports semantic segmentation task.
            train_cfg (dict): The training config.
            pixel_weights (Tensor): Pixels weights.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        seg_logits = self(inputs)
        valid_label_mask = get_valid_label_mask_per_batch(img_metas, self.num_classes)
        losses = self.losses(
            seg_logits, gt_semantic_seg, train_cfg, valid_label_mask=valid_label_mask, pixel_weights=pixel_weights
        )

        if return_logits:
            logits = self.forward_output if self.forward_output is not None else seg_logits
            return losses, logits
        return losses

    @force_fp32(apply_to=("seg_logit",))
    def losses(self, seg_logit, seg_label, train_cfg, valid_label_mask, pixel_weights=None):
        """Compute segmentation loss."""

        loss = dict()

        seg_logit = resize(
            input=seg_logit,
            size=seg_label.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        seg_label = seg_label.squeeze(1)

        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        out_losses = dict()
        for loss_idx, loss_module in enumerate(losses_decode):
            loss_value, loss_meta = loss_module(seg_logit, seg_label, valid_label_mask, pixel_weights=pixel_weights)

            loss_name = loss_module.name + f"-{loss_idx}"
            out_losses[loss_name] = loss_value
            loss.update(add_prefix(loss_meta, loss_name))

        if self.enable_loss_equalizer and len(losses_decode) > 1:
            out_losses = self.loss_equalizer.reweight(out_losses)

        for loss_name, loss_value in out_losses.items():
            loss[loss_name] = loss_value

        loss["loss_seg"] = sum(out_losses.values())
        loss["acc_seg"] = accuracy(seg_logit, seg_label)

        return loss



from copy import deepcopy

import pytest
import torch

from otx.algorithms.common.adapters.mmcv.configs.backbones.lite_hrnet_18 import (
    model as model_cfg,
)
from otx.mpa.modules.models.backbones.litehrnet import (
    LiteHRNet,
    NeighbourSupport,
    SpatialWeightingV2,
    StemV2,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSpatialWeightingV2:
    @e2e_pytest_unit
    def test_forward(self):
        swv2 = SpatialWeightingV2(channels=32)
        assert swv2 is not None

        inputs = torch.randn(1, 32, 32, 32)
        outputs = swv2(inputs)
        assert outputs is not None


class TestStemV2:
    def setup_method(self):
        self.stemv2 = StemV2(in_channels=32, stem_channels=32, out_channels=32, expand_ratio=1)

    @e2e_pytest_unit
    def test_init(self):
        stemv2_extra_stride = StemV2(
            in_channels=32, stem_channels=32, out_channels=32, expand_ratio=1, extra_stride=True
        )
        assert stemv2_extra_stride is not None

        stemv2_input_norm = StemV2(in_channels=32, stem_channels=32, out_channels=32, expand_ratio=1, input_norm=True)
        assert stemv2_input_norm is not None

    @e2e_pytest_unit
    def test_forward(self):
        inputs = torch.randn(1, 32, 32, 32)
        outputs = self.stemv2(inputs)
        assert outputs is not None


class TestNeighbourSupport:
    @e2e_pytest_unit
    def test_forward(self):
        neighbour_support = NeighbourSupport(channels=32)
        assert neighbour_support is not None

        inputs = torch.randn(1, 32, 32, 32)
        outputs = neighbour_support(inputs)
        assert outputs is not None


class TestLiteHRNet:
    def setup_method(self):
        self.extra = model_cfg["backbone"]["extra"]
        self.model = LiteHRNet(extra=self.extra)

    @e2e_pytest_unit
    def test_init(self):
        extra = deepcopy(self.extra)
        extra["out_modules"]["conv"]["enable"] = True
        model = LiteHRNet(extra=extra)
        assert model is not None

        extra["add_stem_features"] = True
        model = LiteHRNet(extra=extra)
        assert model is not None

        extra["stages_spec"]["module_type"] = ("NAIVE", "NAIVE", "NAIVE")
        extra["stages_spec"]["weighting_module_version"] = "v2"
        model = LiteHRNet(extra=extra)
        assert model is not None

    @e2e_pytest_unit
    def test_init_weights(self):
        self.model.init_weights()

        with pytest.raises(TypeError, match="pretrained must be a str or None"):
            self.model.init_weights(0)

    @e2e_pytest_unit
    def test_forward(self):
        self.model.train()
        inputs = torch.randn((1, 3, 224, 224))
        outputs = self.model(inputs)
        assert outputs is not None

        extra = deepcopy(self.extra)
        extra["stages_spec"]["module_type"] = ("NAIVE", "NAIVE", "NAIVE")
        extra["stages_spec"]["weighting_module_version"] = "v2"
        model = LiteHRNet(extra=extra)
        outputs = model(inputs)
        assert outputs is not None



"""MMdet model builder."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Optional, Union

import torch
from mmcv.runner import load_checkpoint
from mmcv.utils import Config, ConfigDict, get_logger

from otx.mpa.utils.logger import LEVEL

logger = get_logger("mmdet")


def build_detector(
    config: Config,
    train_cfg: Optional[Union[Config, ConfigDict]] = None,
    test_cfg: Optional[Union[Config, ConfigDict]] = None,
    checkpoint: Optional[str] = None,
    device: Union[str, torch.device] = "cpu",
    cfg_options: Optional[Union[Config, ConfigDict]] = None,
    from_scratch: bool = False,
) -> torch.nn.Module:
    """A builder function for mmdet model.

    Creates a model, based on the configuration in config.
    Note that this function updates 'load_from' attribute of 'config'.
    """

    from mmdet.models import build_detector as origin_build_detector

    if cfg_options is not None:
        config.merge_from_dict(cfg_options)

    model_cfg = deepcopy(config.model)
    model = origin_build_detector(model_cfg, train_cfg=train_cfg, test_cfg=test_cfg)
    logger.setLevel("WARNING")
    model.init_weights()
    logger.setLevel(LEVEL)
    model = model.to(device)

    checkpoint = checkpoint if checkpoint else config.pop("load_from", None)
    if checkpoint is not None and not from_scratch:
        load_checkpoint(model, checkpoint, map_location=device)
    config.load_from = checkpoint

    return model



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from contextlib import nullcontext

import torch
from mmcv.runner import wrap_fp16_model
from mmcv.utils import Config, ConfigDict
from mmseg.datasets import build_dataloader as mmseg_build_dataloader
from mmseg.datasets import build_dataset as mmseg_build_dataset

from otx.algorithms.common.adapters.mmcv.utils import (
    build_data_parallel,
    build_dataloader,
    build_dataset,
)
from otx.mpa.modules.hooks.recording_forward_hooks import FeatureVectorHook
from otx.mpa.registry import STAGES
from otx.mpa.stage import Stage
from otx.mpa.utils.logger import get_logger

from .stage import SegStage

logger = get_logger()


@STAGES.register_module()
class SegInferrer(SegStage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset = None

    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run inference stage for segmentation

        - Configuration
        - Environment setup
        - Run inference via MMSegmentation -> MMCV
        """
        self._init_logger()
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("infer!")

        model_builder = kwargs.get("model_builder", None)
        dump_features = kwargs.get("dump_features", False)
        outputs = self.infer(
            cfg,
            model_builder=model_builder,
            dump_features=dump_features,
        )

        return dict(outputs=outputs)

    def infer(self, cfg, model_builder=None, dump_features=False):
        # TODO: distributed inference

        data_cfg = cfg.data.test.copy()

        # Input source
        input_source = cfg.get("input_source", "test")
        logger.info(f"Inferring on input source: data.{input_source}")
        if input_source == "train":
            src_data_cfg = Stage.get_data_cfg(cfg, "train")
        else:
            src_data_cfg = cfg.data[input_source]

        if "classes" in src_data_cfg:
            data_cfg.classes = src_data_cfg.classes

        data_cfg = Config(
            ConfigDict(
                data=ConfigDict(
                    samples_per_gpu=cfg.data.get("samples_per_gpu", 1),
                    workers_per_gpu=cfg.data.get("workers_per_gpu", 0),
                    test=data_cfg,
                    test_dataloader=cfg.data.get("test_dataloader", {}).copy(),
                ),
                gpu_ids=cfg.gpu_ids,
                seed=cfg.get("seed", None),
                model_task=cfg.model_task,
            )
        )
        self.configure_samples_per_gpu(data_cfg, "test", distributed=False)
        self.configure_compat_cfg(data_cfg)
        samples_per_gpu = data_cfg.data.test_dataloader.get("samples_per_gpu", 1)
        if samples_per_gpu > 1:
            # Replace 'ImageToTensor' to 'DefaultFormatBundle'
            data_cfg.data.test.pipeline = replace_ImageToTensor(data_cfg.data.test.pipeline)

        # Data loader
        self.dataset = build_dataset(data_cfg, "test", mmseg_build_dataset)
        test_dataloader = build_dataloader(
            self.dataset,
            data_cfg,
            "test",
            mmseg_build_dataloader,
            distributed=False,
            # segmentor does not support various sized batch images
            samples_per_gpu=1,
        )

        # Target classes
        if "task_adapt" in cfg:
            target_classes = cfg.task_adapt.final
            if len(target_classes) < 1:
                raise KeyError(
                    f"target_classes={target_classes} is empty check the metadata from model ckpt or recipe "
                    "configuration"
                )
        else:
            target_classes = self.dataset.CLASSES

        # Model
        cfg.model.pretrained = None
        if cfg.model.get("neck"):
            if isinstance(cfg.model.neck, list):
                for neck_cfg in cfg.model.neck:
                    if neck_cfg.get("rfp_backbone"):
                        if neck_cfg.rfp_backbone.get("pretrained"):
                            neck_cfg.rfp_backbone.pretrained = None
            elif cfg.model.neck.get("rfp_backbone"):
                if cfg.model.neck.rfp_backbone.get("pretrained"):
                    cfg.model.neck.rfp_backbone.pretrained = None
        cfg.model.test_cfg.return_repr_vector = True
        model = self.build_model(cfg, model_builder, fp16=cfg.get("fp16", False))
        model.CLASSES = target_classes
        model.eval()
        feature_model = self._get_feature_module(model)
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        self.set_inference_progress_callback(model, cfg)

        eval_predictions = []
        feature_vectors = []
        with FeatureVectorHook(feature_model) if dump_features else nullcontext() as fhook:
            for data in test_dataloader:
                with torch.no_grad():
                    result = model(return_loss=False, output_logits=True, **data)
                eval_predictions.append(result)
            feature_vectors = fhook.records if dump_features else [None] * len(self.dataset)

        assert len(eval_predictions) == len(feature_vectors), (
            "Number of elements should be the same, however, number of outputs are ",
            f"{len(eval_predictions)} and {len(feature_vectors)}",
        )

        outputs = dict(
            classes=target_classes,
            eval_predictions=eval_predictions,
            feature_vectors=feature_vectors,
        )
        return outputs


import copy  # noqa: E402
import warnings  # noqa: E402


def replace_ImageToTensor(pipelines):
    pipelines = copy.deepcopy(pipelines)
    for i, pipeline in enumerate(pipelines):
        if pipeline["type"] == "MultiScaleFlipAug":
            assert "transforms" in pipeline
            pipeline["transforms"] = replace_ImageToTensor(pipeline["transforms"])
        elif pipeline["type"] == "ImageToTensor":
            warnings.warn(
                '"ImageToTensor" pipeline is replaced by '
                '"DefaultFormatBundle" for batch inference. It is '
                "recommended to manually replace it in the test "
                "data pipeline in your config file.",
                UserWarning,
            )
            pipelines[i] = {"type": "DefaultFormatBundle"}
    return pipelines



import pytest
import torch

from otx.mpa.modules.models.classifiers.semisl_classifier import (
    SAMImageClassifier,
    SemiSLClassifier,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSemiSLClassifier:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch.object(SAMImageClassifier, "__init__", return_value=None)
        self.semisl_classifier = SemiSLClassifier()
        self.kwargs = dict()

    @e2e_pytest_unit
    def test_forward_train(self, mocker):
        img = torch.rand(1, 3, 224, 224)
        self.kwargs["gt_label"] = torch.rand(1, 1)
        self.kwargs["extra_0"] = {"img": torch.rand(1, 3, 224, 224), "img_strong": torch.rand(1, 3, 224, 224)}
        self.semisl_classifier.extract_feat = mocker.MagicMock()
        self.semisl_classifier.head = mocker.MagicMock()
        losses = self.semisl_classifier.forward_train(img, **self.kwargs)

        assert losses is not None



"""Unit Tests for the MPA Dataset Pipelines Transforms - Two Crop."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from mmcls.datasets.builder import PIPELINES
from mmcls.datasets.pipelines import Compose
from mmcv.utils import build_from_cfg

from otx.mpa.modules.datasets.pipelines.transforms.twocrop_transform import (
    TwoCropTransform,
)


def test_TwoCropTransform() -> None:
    """Test the TwoCropTransform instance."""
    # Data to be transformed
    data = {}
    data["img"] = np.ones((224, 224, 3), dtype=np.uint8)
    data["gt_label"] = 0

    # Pipeline to be used for transformation
    pipeline = [
        dict(type="Resize", size=(256, 256)),
        dict(type="RandomCrop", size=(224, 224)),
        dict(
            type="Normalize",
            mean=[123.675, 116.28, 103.53],
            std=[58.395, 57.12, 57.375],
        ),
        dict(type="ToTensor", keys=["img"]),
    ]

    # Create TwoCropTransform instance
    transform = TwoCropTransform(pipeline)
    transform_pipeline = Compose([build_from_cfg(p, PIPELINES) for p in pipeline])
    transform.pipeline1 = transform_pipeline
    transform.pipeline2 = transform_pipeline

    # Test the TwoCropTransform instance
    transformed_data = transform(data)
    assert isinstance(transformed_data, dict)
    assert transformed_data["img"].shape == (2, 224, 224, 3)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.api.configuration.default_model_parameters import DefaultModelParameters
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestModelConfiguration:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_model_configuration(self):
        mc = DefaultModelParameters()
        assert hasattr(mc, "learning_parameters")

        epoch_default = mc.learning_parameters.get_metadata("epochs")["default_value"]
        batch_size_default = mc.learning_parameters.get_metadata("batch_size")["default_value"]

        assert mc.learning_parameters.epochs == epoch_default
        assert mc.learning_parameters.batch_size == batch_size_default

        mc.learning_parameters.epochs = epoch_default + 5
        mc.learning_parameters.batch_size = batch_size_default + 4

        assert mc.learning_parameters.batch_size == batch_size_default + 4
        assert mc.learning_parameters.epochs == epoch_default + 5



"""Configurable parameter conversion between OTX and Anomalib."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pathlib import Path
from typing import Union

import anomalib
from anomalib.config.config import get_configurable_parameters
from omegaconf import DictConfig, ListConfig

from otx.api.configuration.configurable_parameters import ConfigurableParameters


def get_anomalib_config(task_name: str, otx_config: ConfigurableParameters) -> Union[DictConfig, ListConfig]:
    """Get anomalib configuration.

    Create an anomalib config object that matches the values specified in the
    OTX config.

    Args:
        otx_config: ConfigurableParameters: OTX config object parsed from
            configuration.yaml file

    Returns:
        Anomalib config object for the specified model type with overwritten
        default values.
    """
    config_path = Path(anomalib.__file__).parent / "models" / task_name.lower() / "config.yaml"
    anomalib_config = get_configurable_parameters(model_name=task_name.lower(), config_path=config_path)
    # TODO: remove this hard coding of the config location
    if anomalib_config.model.name == "draem":
        anomalib_config.dataset.transform_config.train = (
            f"otx/algorithms/adapters/anomalib/configs/{task_name.lower()}/draem/transform_config.yaml"
        )
        anomalib_config.dataset.transform_config.val = (
            f"otx/algorithms/adapters/anomalib/configs/{task_name.lower()}/draem/transform_config.yaml"
        )
    else:
        anomalib_config.dataset.transform_config.train = None
        anomalib_config.dataset.transform_config.val = None
    update_anomalib_config(anomalib_config, otx_config)
    return anomalib_config


def _anomalib_config_mapper(anomalib_config: Union[DictConfig, ListConfig], otx_config: ConfigurableParameters):
    """Return mapping from learning parameters to anomalib parameters.

    Args:
        anomalib_config: DictConfig: Anomalib config object
        otx_config: ConfigurableParameters: OTX config object parsed from configuration.yaml file
    """
    parameters = otx_config.parameters
    groups = otx_config.groups
    for name in parameters:
        if name == "train_batch_size":
            anomalib_config.dataset["train_batch_size"] = getattr(otx_config, "train_batch_size")
        elif name == "max_epochs":
            anomalib_config.trainer["max_epochs"] = getattr(otx_config, "max_epochs")
        else:
            assert name in anomalib_config.model.keys(), f"Parameter {name} not present in anomalib config."
            sc_value = getattr(otx_config, name)
            sc_value = sc_value.value if hasattr(sc_value, "value") else sc_value
            anomalib_config.model[name] = sc_value
    for group in groups:
        update_anomalib_config(anomalib_config.model[group], getattr(otx_config, group))


def update_anomalib_config(anomalib_config: Union[DictConfig, ListConfig], otx_config: ConfigurableParameters):
    """Update anomalib configuration.

    Overwrite the default parameter values in the anomalib config with the
    values specified in the OTX config. The function is recursively called for
    each parameter group present in the OTX config.

    Args:
        anomalib_config: DictConfig: Anomalib config object
        otx_config: ConfigurableParameters: OTX config object parsed from
            configuration.yaml file
    """
    for param in otx_config.parameters:
        assert param in anomalib_config.keys(), f"Parameter {param} not present in anomalib config."
        sc_value = getattr(otx_config, param)
        sc_value = sc_value.value if hasattr(sc_value, "value") else sc_value
        anomalib_config[param] = sc_value
    for group in otx_config.groups:
        # Since pot_parameters and nncf_optimization are specific to OTX
        if group == "learning_parameters":
            _anomalib_config_mapper(anomalib_config, getattr(otx_config, "learning_parameters"))
        elif group not in ["pot_parameters", "nncf_optimization"]:
            update_anomalib_config(anomalib_config[group], getattr(otx_config, group))



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn


class MixLossMixin(object):
    def forward_train(self, img, img_metas, gt_semantic_seg, aux_img=None, **kwargs):
        """Forward function for training.

        Args:
            img (Tensor): Input images.
            img_metas (list[dict]): List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                `mmseg/datasets/pipelines/formatting.py:Collect`.
            gt_semantic_seg (Tensor): Semantic segmentation masks
                used if the architecture supports semantic segmentation task.
            aux_img (Tensor): Auxiliary images.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        if aux_img is not None:
            mix_loss_enabled = False
            mix_loss_cfg = self.train_cfg.get("mix_loss", None)
            if mix_loss_cfg is not None:
                mix_loss_enabled = mix_loss_cfg.get("enable", False)
            if mix_loss_enabled:
                self.train_cfg.mix_loss.enable = mix_loss_enabled

        if self.train_cfg.mix_loss.enable:
            img = torch.cat([img, aux_img], dim=0)
            gt_semantic_seg = torch.cat([gt_semantic_seg, gt_semantic_seg], dim=0)

        return super().forward_train(img, img_metas, gt_semantic_seg, **kwargs)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass

import torch

from .builder import OPS
from .op import Attribute, Operation

_torch_to_ov = {
    torch.uint8: ["u1", "u4", "u8"],
    torch.int8: ["i4", "i8"],
    torch.int16: ["i16"],
    torch.int32: ["i32"],
    torch.int64: ["i64"],
    torch.float16: ["f16"],
    torch.float32: ["f32"],
    torch.bool: ["boolean"],
}

_ov_to_torch = {
    "u1": torch.uint8,  # no type in torch
    "u4": torch.uint8,  # no type in torch
    "u8": torch.uint8,
    "u32": torch.int32,  # no type in torch
    "u64": torch.int64,  # no type in torch
    "i4": torch.int8,  # no type in torch
    "i8": torch.int8,
    "i16": torch.int16,
    "i32": torch.int32,
    "i64": torch.int64,
    "f16": torch.float16,
    "f32": torch.float32,
    "boolean": torch.bool,
}


@dataclass
class ConvertV0Attribute(Attribute):
    destination_type: str


@OPS.register()
class ConvertV0(Operation[ConvertV0Attribute]):
    TYPE = "Convert"
    VERSION = 0
    ATTRIBUTE_FACTORY = ConvertV0Attribute

    @staticmethod
    def convert_ov_type(ov_type):
        if ov_type not in _ov_to_torch:
            raise NotImplementedError
        return _ov_to_torch[ov_type]

    @staticmethod
    def convert_torch_type(torch_type):
        if torch_type not in _torch_to_ov:
            raise NotImplementedError
        return _torch_to_ov[torch_type][-1]

    def forward(self, input):
        return input.type(self.convert_ov_type(self.attrs.destination_type))



"""Helper functions for computing metrics.

This module contains the helper functions which can be called directly by algorithm implementers to obtain the metrics.
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.accuracy import Accuracy
from otx.api.usecases.evaluation.anomaly_metrics import (
    AnomalyDetectionScores,
    AnomalySegmentationScores,
)
from otx.api.usecases.evaluation.averaging import MetricAverageMethod
from otx.api.usecases.evaluation.dice import DiceAverage
from otx.api.usecases.evaluation.f_measure import FMeasure


class MetricsHelper:
    """Contains metrics computation functions.

    TODO: subject for refactoring.
    """

    @staticmethod
    def compute_f_measure(
        resultset: ResultSetEntity,
        vary_confidence_threshold: bool = False,
        vary_nms_threshold: bool = False,
        cross_class_nms: bool = False,
    ) -> FMeasure:
        """Compute the F-Measure on a resultset given some parameters.

        Args:
            resultset: The resultset used to compute f-measure
            vary_confidence_threshold: Flag specifying whether f-measure
                shall be computed for different confidence threshold
                values
            vary_nms_threshold: Flag specifying whether f-measure shall
                be computed for different NMS threshold values
            cross_class_nms: Whether non-max suppression should be
                applied cross-class

        Returns:
            FMeasure object
        """
        return FMeasure(resultset, vary_confidence_threshold, vary_nms_threshold, cross_class_nms)

    @staticmethod
    def compute_dice_averaged_over_pixels(
        resultset: ResultSetEntity,
        average: MetricAverageMethod = MetricAverageMethod.MACRO,
    ) -> DiceAverage:
        """Compute the Dice average on a resultset, averaged over the pixels.

        Args:
            resultset: The resultset used to compute the Dice average
            average: The averaging method, either MICRO or MACRO

        Returns:
            DiceAverage object
        """
        return DiceAverage(resultset=resultset, average=average)

    @staticmethod
    def compute_accuracy(
        resultset: ResultSetEntity,
        average: MetricAverageMethod = MetricAverageMethod.MICRO,
    ) -> Accuracy:
        """Compute the Accuracy on a resultset, averaged over the different label groups.

        Args:
            resultset: The resultset used to compute the accuracy
            average: The averaging method, either MICRO or MACRO

        Returns:
            Accuracy object
        """
        return Accuracy(resultset=resultset, average=average)

    @staticmethod
    def compute_anomaly_segmentation_scores(
        resultset: ResultSetEntity,
    ) -> AnomalySegmentationScores:
        """Compute the anomaly localization performance metrics on an anomaly segmentation resultset.

        Args:
            resultset: The resultset used to compute the metrics

        Returns:
            AnomalyLocalizationScores object
        """
        return AnomalySegmentationScores(resultset)

    @staticmethod
    def compute_anomaly_detection_scores(
        resultset: ResultSetEntity,
    ) -> AnomalyDetectionScores:
        """Compute the anomaly localization performance metrics on an anomaly detection resultset.

        Args:
            resultset: The resultset used to compute the metrics

        Returns:
            AnomalyLocalizationScores object
        """
        return AnomalyDetectionScores(resultset)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools

from mmdet.models.builder import DETECTORS
from mmdet.models.detectors.two_stage import TwoStageDetector

from otx.mpa.modules.utils.task_adapt import map_class_names
from otx.mpa.utils.logger import get_logger

from .l2sp_detector_mixin import L2SPDetectorMixin
from .sam_detector_mixin import SAMDetectorMixin

logger = get_logger()


@DETECTORS.register_module()
class CustomTwoStageDetector(SAMDetectorMixin, L2SPDetectorMixin, TwoStageDetector):
    """SAM optimizer & L2SP regularizer enabled custom 2-stage detector"""

    def __init__(self, *args, task_adapt=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Hook for class-sensitive weight loading
        if task_adapt:
            self._register_load_state_dict_pre_hook(
                functools.partial(
                    self.load_state_dict_pre_hook,
                    self,  # model
                    task_adapt["dst_classes"],  # model_classes
                    task_adapt["src_classes"],  # chkpt_classes
                )
            )

    def forward_train(self, img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=None, **kwargs):
        return super().forward_train(img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=gt_bboxes_ignore)

    @staticmethod
    def load_state_dict_pre_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading"""
        logger.info(f"----------------- CustomTwoStageDetector.load_state_dict_pre_hook() called w/ prefix: {prefix}")

        # Dst to src mapping index
        model_dict = model.state_dict()
        model_classes = list(model_classes)
        chkpt_classes = list(chkpt_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes} ({model2chkpt})")

        # List of class-relevant params & their row-stride
        param_strides = {
            "roi_head.bbox_head.fc_cls.weight": 1,
            "roi_head.bbox_head.fc_cls.bias": 1,
            "roi_head.bbox_head.fc_reg.weight": 4,  # 4 rows (bbox) for each class
            "roi_head.bbox_head.fc_reg.bias": 4,
        }

        for model_name, stride in param_strides.items():
            chkpt_name = prefix + model_name
            if model_name not in model_dict or chkpt_name not in chkpt_dict:
                logger.info(f"Skipping weight copy: {chkpt_name}")
                continue

            # Mix weights
            model_param = model_dict[model_name].clone()
            chkpt_param = chkpt_dict[chkpt_name]
            for m, c in enumerate(model2chkpt):
                if c >= 0:
                    # Copying only matched weight rows
                    model_param[(m) * stride : (m + 1) * stride].copy_(chkpt_param[(c) * stride : (c + 1) * stride])
            if model_param.shape[0] > len(model_classes * stride):  # BG class
                c = len(chkpt_classes)
                m = len(model_classes)
                model_param[(m) * stride : (m + 1) * stride].copy_(chkpt_param[(c) * stride : (c + 1) * stride])

            # Replace checkpoint weight by mixed weights
            chkpt_dict[chkpt_name] = model_param



"""Unit Test for otx.algorithms.action.adapters.mmaction.data.cls_dataset."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.action.adapters.mmaction.data.cls_dataset import OTXActionClsDataset
from otx.algorithms.action.adapters.mmaction.data.pipelines import RawFrameDecode
from otx.algorithms.action.configs.classification.x3d.data_pipeline import (
    train_pipeline,
)
from otx.api.entities.label import Domain
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    MockPipeline,
    generate_action_cls_otx_dataset,
    generate_labels,
)


class TestOTXActionClsDataset:
    """Test OTXActionClsDataset class.

    1. Check _DataInfoProxy
    <Steps>
        1. Create otx_dataset, labels
        2. Check len(_DataInfoProxy)
    2. Check data pipelines
    3. Check "__len__" function
    4. Check "prepare_train_frames" function
    5. Check "prepare_test_frames" function
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.video_len = 3
        self.frame_len = 3
        self.labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.otx_dataset = generate_action_cls_otx_dataset(self.video_len, self.frame_len, self.labels)
        self.pipeline = train_pipeline

    @e2e_pytest_unit
    def test_DataInfoProxy(self) -> None:
        """Test _DataInfoProxy Class."""

        proxy = OTXActionClsDataset._DataInfoProxy(self.otx_dataset, self.labels, modality="RGB")
        sample = proxy[0]
        assert len(proxy) == self.video_len
        assert "total_frames" in sample
        assert "start_index" in sample
        assert "label" in sample

    @e2e_pytest_unit
    def test_pipeline(self) -> None:
        """Test RawFrameDecode transform contains otx_dataset."""

        dataset = OTXActionClsDataset(self.otx_dataset, self.labels, self.pipeline)
        for transform in dataset.pipeline.transforms:
            if isinstance(transform, RawFrameDecode):
                assert transform.otx_dataset == self.otx_dataset

    @e2e_pytest_unit
    def test_len(self) -> None:
        """Test dataset length is same with video_len."""

        dataset = OTXActionClsDataset(self.otx_dataset, self.labels, self.pipeline)
        assert len(dataset) == self.video_len

    @e2e_pytest_unit
    def test_prepare_train_frames(self) -> None:
        """Test prepare_train_frames function."""

        dataset = OTXActionClsDataset(self.otx_dataset, self.labels, self.pipeline)
        dataset.pipeline = MockPipeline()
        sample = dataset.prepare_train_frames(0)
        assert "total_frames" in sample
        assert "start_index" in sample
        assert "label" in sample

    @e2e_pytest_unit
    def test_prepare_test_frames(self) -> None:
        """Test prepare_test_frames function."""

        dataset = OTXActionClsDataset(self.otx_dataset, self.labels, self.pipeline)
        dataset.pipeline = MockPipeline()
        sample = dataset.prepare_test_frames(0)
        assert "total_frames" in sample
        assert "start_index" in sample
        assert "label" in sample



# Copyright (c) 2020-2021 The MMSegmentation Authors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn

from ..utils import IterativeAggregator, IterativeConcatAggregator


class AggregatorMixin(nn.Module):
    def __init__(
        self,
        *args,
        enable_aggregator=False,
        aggregator_min_channels=None,
        aggregator_merge_norm=None,
        aggregator_use_concat=False,
        **kwargs
    ):

        in_channels = kwargs.get("in_channels")
        in_index = kwargs.get("in_index")
        norm_cfg = kwargs.get("norm_cfg")
        conv_cfg = kwargs.get("conv_cfg")
        input_transform = kwargs.get("input_transform")

        aggregator = None
        if enable_aggregator:
            assert isinstance(in_channels, (tuple, list))
            assert len(in_channels) > 1

            aggregator = IterativeAggregator(
                in_channels=in_channels,
                min_channels=aggregator_min_channels,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                merge_norm=aggregator_merge_norm,
                use_concat=aggregator_use_concat,
            )

            aggregator_min_channels = aggregator_min_channels if aggregator_min_channels is not None else 0
            # change arguments temporarily
            kwargs["in_channels"] = max(in_channels[0], aggregator_min_channels)
            kwargs["input_transform"] = None
            if in_index is not None:
                kwargs["in_index"] = in_index[0]

        super(AggregatorMixin, self).__init__(*args, **kwargs)

        self.aggregator = aggregator
        # re-define variables
        self.in_channels = in_channels
        self.input_transform = input_transform
        self.in_index = in_index

    def _transform_inputs(self, inputs):
        inputs = super()._transform_inputs(inputs)
        if self.aggregator is not None:
            inputs = self.aggregator(inputs)[0]
        return inputs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import datetime
import json

from mmcv.utils import Registry

from otx.mpa.utils.logger import get_logger

logger = get_logger()
WORKFLOW_HOOKS = Registry("workflow_hooks")


def build_workflow_hook(config, *args, **kwargs):
    logger.info(f"called build_workflow_hook({config})")
    whook_type = config.pop("type")
    # event = config.pop('event')
    if whook_type not in WORKFLOW_HOOKS:
        raise KeyError(f"not supported workflow hook type {whook_type}")
    else:
        whook_cls = WORKFLOW_HOOKS.get(whook_type)
    return whook_cls(*args, **kwargs, **config)


class WorkflowHook:
    def __init__(self, name):
        self.name = name

    def before_workflow(self, workflow, idx=-1, results=None):
        pass

    def after_workflow(self, workflow, idx=-1, results=None):
        pass

    def before_stage(self, workflow, idx, results=None):
        pass

    def after_stage(self, workflow, idx, results=None):
        pass


@WORKFLOW_HOOKS.register_module()
class SampleLoggingHook(WorkflowHook):
    def __init__(self, name=__name__, log_level="DEBUG"):
        super(SampleLoggingHook, self).__init__(name)
        self.logging = getattr(logger, log_level.lower())

    def before_stage(self, wf, stage_idx, results):
        self.logging(f"called {self.name}.run()")
        self.logging(f"stage index {stage_idx}, results keys = {results.keys()}")
        result_key = f"{self.name}|{stage_idx}"
        results[result_key] = dict(message=f"this is a sample result of the {__name__} hook")


@WORKFLOW_HOOKS.register_module()
class WFProfileHook(WorkflowHook):
    def __init__(self, name=__name__, output_path=None):
        super(WFProfileHook, self).__init__(name)
        self.output_path = output_path
        self.profile = dict(start=0, end=0, elapsed=0, stages=dict())
        logger.info(f"initialized {__name__}....")

    def before_workflow(self, wf, idx=-1, results=None):
        self.profile["start"] = datetime.datetime.now()

    def after_workflow(self, wf, idx=-1, results=None):
        self.profile["end"] = datetime.datetime.now()
        self.profile["elapsed"] = self.profile["end"] - self.profile["start"]

        str_dumps = json.dumps(self.profile, indent=2, default=str)
        logger.info("** workflow profile results **")
        logger.info(str_dumps)
        if self.output_path is not None:
            with open(self.output_path, "w") as f:
                f.write(str_dumps)

    def before_stage(self, wf, idx=-1, results=None):
        stages = self.profile.get("stages")
        stages[f"{idx}"] = {}
        stages[f"{idx}"]["start"] = datetime.datetime.now()

    def after_stage(self, wf, idx=-1, results=None):
        stages = self.profile.get("stages")
        stages[f"{idx}"]["end"] = datetime.datetime.now()
        stages[f"{idx}"]["elapsed"] = stages[f"{idx}"]["end"] - stages[f"{idx}"]["start"]


@WORKFLOW_HOOKS.register_module()
class AfterStageWFHook(WorkflowHook):
    def __init__(self, name, stage_cfg_updated_callback):
        self.callback = stage_cfg_updated_callback
        super().__init__(name)

    def after_stage(self, workflow, idx, results=None):
        logger.info(f"{__name__}: called after_stage()")
        name = copy.deepcopy(workflow.stages[idx].name)
        cfg = copy.deepcopy(workflow.stages[idx].cfg)
        self.callback(name, cfg)



"""
Requirements module.
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


class Requirements:
    """
    Dummy requirement.
    """

    REQ_1 = "Dummy requirement"



# Copyright (c) 2018-2020 Open-MMLab.
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch


def channel_shuffle(x, groups):
    """Channel Shuffle operation.

    This function enables cross-group information flow for multiple groups
    convolution layers.

    Args:
        x (Tensor): The input tensor.
        groups (int): The number of groups to divide the input tensor
                      in the channel dimension.

    Returns:
        Tensor: The output tensor after channel shuffle operation.
    """

    batch_size, num_channels, height, width = x.size()
    assert num_channels % groups == 0, "num_channels should be divisible by groups"

    channels_per_group = num_channels // groups

    x = x.view(batch_size, groups, channels_per_group, height, width)
    x = torch.transpose(x, 1, 2).contiguous()
    x = x.view(batch_size, -1, height, width)

    return x



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import json
from pathlib import Path

from otx.algorithms.classification.utils.convert_coco_to_multilabel import (
    coco_to_datumaro_multilabel,
    multilabel_ann_format,
)
from otx.algorithms.classification.utils.labels_utils import get_multihead_class_info
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.classification.test_helper import (
    generate_cls_dataset,
    generate_label_schema,
)


@e2e_pytest_unit
def test_coco_conversion(tmp_dir_path):
    path_to_coco_example_ann_file = (
        Path("tests/assets/car_tree_bug/annotations/instances_train.json").absolute().as_posix()
    )
    path_to_coco_example_image_dir = Path("tests/assets/car_tree_bug/images/train").absolute().as_posix()
    output_path = Path(tmp_dir_path) / "annotations.json"
    coco_to_datumaro_multilabel(
        path_to_coco_example_ann_file, path_to_coco_example_image_dir, output_path, test_mode=True
    )
    assert Path.exists(output_path)
    with open(output_path, "r") as f:
        coco_ann = json.load(f)
        for key in multilabel_ann_format:
            assert key in coco_ann.keys()
        assert len(coco_ann["items"]) > 0


@e2e_pytest_unit
def test_get_multihead_class_info():
    hierarchical_dataset = generate_cls_dataset(hierarchical=True)
    label_schema = generate_label_schema(hierarchical_dataset.get_labels(), multilabel=False, hierarchical=True)
    class_info = get_multihead_class_info(label_schema)
    assert (
        len(class_info["label_to_idx"])
        == len(class_info["class_to_group_idx"])
        == len(hierarchical_dataset.get_labels())
    )



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import subprocess
import sys
import time


def __mo_check_requirements(framework="onnx"):
    from mo.utils.versions_checker import check_requirements as mo_check_requirements

    return mo_check_requirements(framework)


def check_requirements_with_version(target, framework=None):
    from mo.utils.version import get_version as mo_get_version

    mo_version = mo_get_version()
    print("mo vesion =", mo_version)
    # # TODO: model optimizer version in OpenVINO 2019.3.3 is 2019.3.0
    # # it will be changed before official release to meet with requirement of VAS
    # required_mo_version = '2019.3.0'
    # if target.lower() == 'fr':
    #     required_mo_version = '2019.1.1'
    # if not mo_version.startswith(required_mo_version):
    #     print('current MO version: {} but required version: {}'.format(mo_version, required_mo_version))
    #     return False
    err_code = __mo_check_requirements(framework)
    if err_code:
        print("mo_check_requriements returns: {}".format(err_code))
        return False
    return True


def check_requirements(framework="onnx"):
    err_code = __mo_check_requirements(framework)
    if err_code:
        print("mo_check_requriements returns: {}".format(err_code))
        return False
    return True


def __mo_main_wrapper(argv, framework=None):
    """MO main wrapper.

    See: openvino_2019.1.094/deployment_tools/model_optimizer/mo/main.py
    """
    # substitute the value of sys.argv with a proper cli arguments to call the
    # OpenVINO model optimizer's main function
    from mo.main import main as mo_main

    old_argv = sys.argv
    sys.argv = [sys.argv[0]] + argv

    # run mo
    from mo.utils.cli_parser import get_all_cli_parser

    ret = mo_main(get_all_cli_parser(), framework)

    # restore sys.argv
    sys.argv = old_argv

    return ret


MO_LOG_LEVELS = ["CRITICAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG", "NOTSET"]

MO_ARGS = [
    "input_model",
    "input_shape",
    "input",
    "mean_values",
    "scale",
    "model_name",
    "log_level",
    "compress_to_fp16",
    "scale_values",
    "disable_fusing",
    "transformations_config",
    "reverse_input_channels",
    "output_dir",
]


def generate_ir(output_path, model_path, silent, save_xml=True, **mo_kwargs):
    # parse kwargs for the model optimizer
    mo_args = []
    for key, value in mo_kwargs.items():
        if key not in MO_ARGS:
            return -1, "Not supported argument: {}".format(key)
        if value is not None:
            mo_args.append("--{}={}".format(key, value))
        else:
            mo_args.append("--{}".format(key))

    is_output_dir_provided = False
    for mo_arg in mo_args:
        if mo_arg.startswith("--output_dir"):
            is_output_dir_provided = True
            break
    if not is_output_dir_provided:
        mo_args.append("--output_dir={}".format(model_path))
    print("mo-args: {}".format(mo_args))

    if silent:
        # redirect stdout messages from MO to null device
        devnull = open("/dev/null", "w")
        old_stdout = sys.stdout
        sys.stdout = devnull

    # ret = __mo_main_wrapper(mo_args, None)
    # ret = os.system('mo.py ' + ' '.join(mo_args))
    ret = subprocess.run(["mo"] + mo_args, shell=False).returncode

    if silent:
        # return back stdout
        sys.stdout = old_stdout

    # NOTE: mo returns non zero return code (245) even though it successfully generate IR
    cur_time = time.time()
    time_threshold = 5
    model_name = mo_kwargs.get("model_name", "model")
    if not (
        ret == 245
        and not {f"{model_name}.bin", f"{model_name}.xml"} - set(os.listdir(model_path))
        and (
            os.path.getmtime(os.path.join(model_path, f"{model_name}.bin")) - cur_time < time_threshold
            and os.path.getmtime(os.path.join(model_path, f"{model_name}.xml")) - cur_time < time_threshold
        )
    ):
        err_msg = "Failed to run the model optimizer to convert a model"
        return ret, err_msg

    print("*** Model optimization completed ***")
    # move bin files to workspace

    if not os.path.exists(output_path):
        os.makedirs(output_path)
    os.rename(
        os.path.join(model_path, model_name + ".bin"),
        os.path.join(output_path, model_name + ".bin"),
    )
    if save_xml:
        os.rename(
            os.path.join(model_path, model_name + ".xml"),
            os.path.join(output_path, model_name + ".xml"),
        )

    return 0, "Saved outputs into {}".format(output_path)



"""Data Pipeline of SSD model for Semi-Supervised Learning Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

# This is from otx/mpa/recipes/stages/_base_/data/pipelines/ubt.py
# This could be needed sync with incr-learning's data pipeline
__img_scale = (864, 864)
__img_norm_cfg = dict(mean=[0, 0, 0], std=[255, 255, 255], to_rgb=True)

common_pipeline = [
    dict(
        type="Resize",
        img_scale=__img_scale,
        multiscale_mode="value",
        keep_ratio=False,
    ),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="BranchImage", key_map=dict(img="img0")),
    dict(type="NDArrayToPILImage", keys=["img"]),
    dict(
        type="RandomApply",
        transform_cfgs=[
            dict(
                type="ColorJitter",
                brightness=0.4,
                contrast=0.4,
                saturation=0.4,
                hue=0.1,
            )
        ],
        p=0.8,
    ),
    dict(type="RandomGrayscale", p=0.2),
    dict(
        type="RandomApply",
        transform_cfgs=[
            dict(
                type="RandomGaussianBlur",
                sigma_min=0.1,
                sigma_max=2.0,
            )
        ],
        p=0.5,
    ),
    dict(type="PILImageToNDArray", keys=["img"]),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="Pad", size_divisor=32),
    dict(type="NDArrayToTensor", keys=["img", "img0"]),
    dict(
        type="RandomErasing",
        p=0.7,
        scale=[0.05, 0.2],
        ratio=[0.3, 3.3],
        value="random",
    ),
    dict(
        type="RandomErasing",
        p=0.5,
        scale=[0.02, 0.2],
        ratio=[0.10, 6.0],
        value="random",
    ),
    dict(
        type="RandomErasing",
        p=0.3,
        scale=[0.02, 0.2],
        ratio=[0.05, 8.0],
        value="random",
    ),
]

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="MinIoURandomCrop", min_ious=(0.1, 0.3, 0.5, 0.7, 0.9), min_crop_size=0.3),
    *common_pipeline,
    dict(type="ToTensor", keys=["gt_bboxes", "gt_labels"]),
    dict(
        type="ToDataContainer",
        fields=[
            dict(key="img", stack=True),
            dict(key="img0", stack=True),
            dict(key="gt_bboxes"),
            dict(key="gt_labels"),
        ],
    ),
    dict(
        type="Collect",
        keys=["img", "img0", "gt_bboxes", "gt_labels"],
    ),
]

unlabeled_pipeline = [
    dict(type="LoadImageFromFile"),
    *common_pipeline,
    dict(
        type="ToDataContainer",
        fields=[
            dict(key="img", stack=True),
            dict(key="img0", stack=True),
        ],
    ),
    dict(
        type="Collect",
        keys=[
            "img",
            "img0",
        ],
    ),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_scale,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="Pad", size_divisor=32),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]
data = dict(
    samples_per_gpu=10,
    workers_per_gpu=2,
    train=dict(
        pipeline=train_pipeline,
    ),
    val=dict(
        pipeline=test_pipeline,
    ),
    test=dict(
        pipeline=test_pipeline,
    ),
    unlabeled=dict(
        pipeline=unlabeled_pipeline,
    ),
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field
from typing import Callable, List

import torch
from torch.nn import functional as F

from .builder import OPS
from .op import Attribute, Operation
from .utils import get_torch_padding


@dataclass
class ConvolutionV1Attribute(Attribute):
    strides: List[int]
    pads_begin: List[int]
    pads_end: List[int]
    dilations: List[int]
    auto_pad: str = field(default="explicit")

    def __post_init__(self):
        super().__post_init__()
        valid_auto_pad = ["explicit", "same_upper", "same_Lower", "valid"]
        if self.auto_pad not in valid_auto_pad:
            raise ValueError(f"Invalid auto_pad {self.auto_pad}. " f"It must be one of {valid_auto_pad}.")


@OPS.register()
class ConvolutionV1(Operation[ConvolutionV1Attribute]):
    TYPE = "Convolution"
    VERSION = 1
    ATTRIBUTE_FACTORY = ConvolutionV1Attribute

    def forward(self, input, weight):
        if weight.dim() == 3:
            func = F.conv1d
        elif weight.dim() == 4:
            func = F.conv2d
        elif weight.dim() == 5:
            func = F.conv3d
        else:
            raise NotImplementedError

        padding = get_torch_padding(
            self.attrs.pads_begin,
            self.attrs.pads_end,
            self.attrs.auto_pad,
            list(input.shape[2:]),
            list(weight.shape[2:]),
            self.attrs.strides,
            self.attrs.dilations,
        )
        if isinstance(padding, Callable):
            input = padding(input=input)
            padding = 0

        return func(
            input=input,
            weight=weight,
            bias=None,
            stride=self.attrs.strides,
            padding=padding,
            dilation=self.attrs.dilations,
        )


@dataclass
class GroupConvolutionV1Attribute(ConvolutionV1Attribute):
    pass


@OPS.register()
class GroupConvolutionV1(Operation[GroupConvolutionV1Attribute]):
    TYPE = "GroupConvolution"
    VERSION = 1
    ATTRIBUTE_FACTORY = GroupConvolutionV1Attribute

    def forward(self, input, weight):
        if weight.dim() == 4:
            func = F.conv1d
        elif weight.dim() == 5:
            func = F.conv2d
        elif weight.dim() == 6:
            func = F.conv3d
        else:
            raise NotImplementedError

        n_groups = weight.shape[0]
        # merge groups and out dimension
        weight = weight.view(-1, *weight.shape[2:])

        padding = get_torch_padding(
            self.attrs.pads_begin,
            self.attrs.pads_end,
            self.attrs.auto_pad,
            list(input.shape[2:]),
            list(weight.shape[2:]),
            self.attrs.strides,
            self.attrs.dilations,
        )
        if isinstance(padding, Callable):
            input = padding(input=input)
            padding = 0

        output = func(
            input=input,
            weight=weight,
            bias=None,
            stride=self.attrs.strides,
            padding=padding,
            dilation=self.attrs.dilations,
            groups=n_groups,
        )

        return output



"""Base Configuration of OTX Common Algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from sys import maxsize

from attr import attrs

from otx.api.configuration import ConfigurableEnum, ConfigurableParameters
from otx.api.configuration.elements import (
    ParameterGroup,
    add_parameter_group,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    selectable,
    string_attribute,
)
from otx.api.configuration.model_lifecycle import ModelLifecycle

from .configuration_enums import POTQuantizationPreset


class TrainType(ConfigurableEnum):
    """TrainType for OTX Algorithms."""

    FINETUNE = "FINETUNE"
    SEMISUPERVISED = "SEMISUPERVISED"
    SELFSUPERVISED = "SELFSUPERVISED"
    INCREMENTAL = "INCREMENTAL"
    FUTUREWORK = "FUTUREWORK"


class LearningRateSchedule(ConfigurableEnum):
    """LearningRateSchedule for OTX Algorithms."""

    FIXED = "fixed"
    EXPONENTIAL = "exponential"
    COSINE = "cosine"
    STEP_WISE = "step_wise"
    CYCLIC = "cyclic"
    CUSTOM = "custom"


@attrs
class BaseConfig(ConfigurableParameters):
    """BaseConfig Class for OTX Algorithms."""

    @attrs
    class BaseLearningParameters(ParameterGroup):
        """BaseLearningParameters for OTX Algorithms."""

        batch_size = configurable_integer(
            default_value=5,
            min_value=1,
            max_value=512,
            header="Batch size",
            description="The number of training samples seen in each iteration of training. Increasing thisvalue "
            "improves training time and may make the training more stable. A larger batch size has higher "
            "memory requirements.",
            warning="Increasing this value may cause the system to use more memory than available, "
            "potentially causing out of memory errors, please update with caution.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        unlabeled_batch_size = configurable_integer(
            default_value=5,
            min_value=1,
            max_value=512,
            header="Unlabeled batch size",
            description="The number of unlabeled training samples seen in each iteration of semi-supervised learning."
            "Increasing this value improves training time and may make the training more stable."
            "A larger batch size has higher memory requirements.",
            warning="Increasing this value may cause the system to use more memory than available, "
            "potentially causing out of memory errors, please update with caution.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        num_iters = configurable_integer(
            default_value=1,
            min_value=1,
            max_value=100000,
            header="Number of training iterations",
            description="Increasing this value causes the results to be more robust but training time will be longer.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        learning_rate = configurable_float(
            default_value=0.01,
            min_value=1e-07,
            max_value=1.0,
            header="Learning rate",
            description="Increasing this value will speed up training convergence but might make it unstable.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        learning_rate_warmup_iters = configurable_integer(
            default_value=100,
            min_value=0,
            max_value=10000,
            header="Number of iterations for learning rate warmup",
            description="",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        use_adaptive_interval = configurable_boolean(
            default_value=False,
            header="Use adaptive validation interval",
            description="Depending on the size of iteration per epoch, \
                         adaptively update the validation interval and related values.",
            warning="This will automatically control the patience and interval when early stopping is enabled.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        enable_early_stopping = configurable_boolean(
            default_value=True,
            header="Enable early stopping of the training",
            description="Early exit from training when validation accuracy isn't \
                         changed or decreased for several epochs.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        early_stop_start = configurable_integer(
            default_value=3,
            min_value=0,
            max_value=1000,
            header="Start epoch for early stopping",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        early_stop_patience = configurable_integer(
            default_value=5,
            min_value=0,
            max_value=50,
            header="Patience for early stopping",
            description="Training will stop if the model does not improve within the number of epochs of patience.",
            warning="This is applied exclusively when early stopping is enabled.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        early_stop_iteration_patience = configurable_integer(
            default_value=0,
            min_value=0,
            max_value=1000,
            header="Iteration patience for early stopping",
            description="Training will stop if the model does not improve within the number of iterations of patience. \
                        the model is trained enough with the number of iterations of patience before early stopping.",
            warning="This is applied exclusively when early stopping is enabled.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        num_workers = configurable_integer(
            default_value=0,
            min_value=0,
            max_value=8,
            header="Number of cpu threads to use during batch generation",
            description="Increasing this value might improve training speed however it might cause out of memory "
            "errors. If the number of workers is set to zero, data loading will happen in the main "
            "training thread.",
            affects_outcome_of=ModelLifecycle.NONE,
        )

        num_checkpoints = configurable_integer(
            default_value=5,
            min_value=1,
            max_value=100,
            header="Number of checkpoints that is done during the single training round",
            description="",
            affects_outcome_of=ModelLifecycle.NONE,
        )

        enable_supcon = configurable_boolean(
            default_value=False,
            header="Enable Supervised Contrastive helper loss",
            description="This auxiliar loss might increase robustness and accuracy for small datasets",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    @attrs
    class BasePostprocessing(ParameterGroup):
        """BasePostprocessing for OTX Algorithms."""

        result_based_confidence_threshold = configurable_boolean(
            default_value=True,
            header="Result based confidence threshold",
            description="Confidence threshold is derived from the results",
            affects_outcome_of=ModelLifecycle.INFERENCE,
        )

        confidence_threshold = configurable_float(
            default_value=0.35,
            min_value=0,
            max_value=1,
            header="Confidence threshold",
            description="This threshold only takes effect if the threshold is not set based on the result.",
            affects_outcome_of=ModelLifecycle.INFERENCE,
        )

    @attrs
    class BaseNNCFOptimization(ParameterGroup):
        """BaseNNCFOptimization for OTX Algorithms."""

        enable_quantization = configurable_boolean(
            default_value=True,
            header="Enable quantization algorithm",
            description="Enable quantization algorithm",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        enable_pruning = configurable_boolean(
            default_value=False,
            header="Enable filter pruning algorithm",
            description="Enable filter pruning algorithm",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        pruning_supported = configurable_boolean(
            default_value=False,
            header="Whether filter pruning is supported",
            description="Whether filter pruning is supported",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        maximal_accuracy_degradation = configurable_float(
            default_value=1.0,
            min_value=0.0,
            max_value=100.0,
            header="Maximum accuracy degradation",
            description="The maximal allowed accuracy metric drop",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    @attrs
    class BasePOTParameter(ParameterGroup):
        """BasePOTParameter for OTX Algorithms."""

        stat_subset_size = configurable_integer(
            header="Number of data samples",
            description="Number of data samples used for post-training optimization",
            default_value=300,
            min_value=1,
            max_value=maxsize,
        )

        stat_requests_number = configurable_integer(
            header="Number of requests",
            description="Number of requests during statistics collection",
            default_value=0,
            min_value=0,
            max_value=maxsize,
        )

        preset = selectable(
            default_value=POTQuantizationPreset.PERFORMANCE,
            header="Preset",
            description="Quantization preset that defines quantization scheme",
            editable=True,
            visible_in_ui=True,
        )

    @attrs
    class BaseAlgoBackendParameters(ParameterGroup):
        """BaseAlgoBackendParameters for OTX Algorithms."""

        train_type = selectable(
            default_value=TrainType.INCREMENTAL,
            header="train type",
            description="Training scheme option that determines how to train the model",
            editable=False,
            visible_in_ui=True,
        )

    @attrs
    class BaseTilingParameters(ParameterGroup):
        """BaseTilingParameters for OTX Algorithms."""

        header = string_attribute("Tiling Parameters")
        enable_tiling = configurable_boolean(
            default_value=False,
            header="Enable tiling",
            description="Set to True to allow tiny objects to be better detected.",
            warning="Tiling trades off speed for accuracy as it increases the number of images to be processed.",
            affects_outcome_of=ModelLifecycle.NONE,
        )

        enable_adaptive_params = configurable_boolean(
            default_value=True,
            header="Enable adaptive tiling parameters",
            description="Config tile size and tile overlap adaptively based on annotated dataset statistic",
            warning="",
            affects_outcome_of=ModelLifecycle.NONE,
        )

        tile_size = configurable_integer(
            header="Tile Image Size",
            description="Tile Image Size",
            default_value=400,
            min_value=100,
            max_value=1024,
            affects_outcome_of=ModelLifecycle.NONE,
        )

        tile_overlap = configurable_float(
            header="Tile Overlap",
            description="Overlap between each two neighboring tiles.",
            default_value=0.2,
            min_value=0.0,
            max_value=1.0,
            affects_outcome_of=ModelLifecycle.NONE,
        )

        tile_max_number = configurable_integer(
            header="Max object per image",
            description="Max object per image",
            default_value=1500,
            min_value=1,
            max_value=10000,
            affects_outcome_of=ModelLifecycle.NONE,
        )

    tiling_parameters = add_parameter_group(BaseTilingParameters)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools
import logging
import os
import sys
from typing import Callable

import torch.distributed as dist

# __all__ = ['config_logger', 'get_log_dir', 'get_logger']
__all__ = ["config_logger", "get_log_dir"]

_LOGGING_FORMAT = "%(asctime)s | %(levelname)s : %(message)s"
_LOG_DIR = None
_FILE_HANDLER = None
_CUSTOM_LOG_LEVEL = 31

LEVEL = logging.INFO

logging.addLevelName(_CUSTOM_LOG_LEVEL, "LOG")


def _get_logger():
    logger = logging.getLogger("mpa")
    logger.propagate = False

    def print(message, *args, **kws):
        if logger.isEnabledFor(_CUSTOM_LOG_LEVEL):
            logger.log(_CUSTOM_LOG_LEVEL, message, *args, **kws)

    logger.print = print

    logger.setLevel(LEVEL)
    console = logging.StreamHandler(sys.stdout)
    console.setFormatter(logging.Formatter(_LOGGING_FORMAT))

    logger.addHandler(console)

    return logger


_logger = _get_logger()

# # to expose supported APIs
# _override_methods = ['setLevel', 'addHandler', 'addFilter', 'info',
#                      'warning', 'error', 'critical', 'print']
# for fn in _override_methods:
#     locals()[fn] = getattr(_logger, fn)
#     __all__.append(fn)


def config_logger(log_file, level="WARNING"):
    global _LOG_DIR, _FILE_HANDLER
    if _FILE_HANDLER is not None:
        _logger.removeHandler(_FILE_HANDLER)
        del _FILE_HANDLER

    _LOG_DIR = os.path.dirname(log_file)
    os.makedirs(_LOG_DIR, exist_ok=True)
    file = logging.FileHandler(log_file, mode="w", encoding="utf-8")
    file.setFormatter(logging.Formatter(_LOGGING_FORMAT))
    _FILE_HANDLER = file
    _logger.addHandler(file)
    _logger.setLevel(_get_log_level(level))


def _get_log_level(level):
    # sanity checks
    if level is None:
        return None

    # get level number
    level_number = logging.getLevelName(level.upper())
    if level_number not in [0, 10, 20, 30, 40, 50, _CUSTOM_LOG_LEVEL]:
        msg = "Log level must be one of DEBUG/INFO/WARN/ERROR/CRITICAL/LOG" ", but {} is given.".format(level)
        raise ValueError(msg)

    return level_number


def get_log_dir():
    return _LOG_DIR


class _DummyLogger(logging.Logger):
    def debug(message, *args, **kws):
        pass

    def info(message, *args, **kws):
        pass

    def warning(message, *args, **kws):
        pass

    def critical(message, *args, **kws):
        pass

    def error(message, *args, **kws):
        pass


def local_master_only(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        local_rank = 0
        if dist.is_available() and dist.is_initialized():
            local_rank = int(os.environ["LOCAL_RANK"])
        if local_rank == 0:
            return func(*args, **kwargs)

    return wrapper


# apply decorator @local_master_only to the lower severity logging functions
_logging_methods = ["print", "debug", "info", "warning"]
for fn in _logging_methods:
    setattr(_logger, fn, local_master_only(getattr(_logger, fn)))


def get_logger():
    # if dist.is_available() and dist.is_initialized():
    #     rank = dist.get_rank()
    # else:
    #     rank = 0
    # if rank == 0:
    #     return _logger
    # return _DummyLogger('dummy')
    return _logger



# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import os
import tempfile

import torch.nn as nn
from mmcv.cnn import build_activation_layer, build_norm_layer
from mmcv.runner import get_dist_info

# from ..builder import BACKBONES
from mmdet.models.builder import BACKBONES
from mmdet.utils.logger import get_root_logger
from pytorchcv.model_provider import _models
from pytorchcv.models.model_store import download_model
from torch import distributed
from torch.nn.modules.batchnorm import _BatchNorm

# import types


def replace_activation(model, activation_cfg):
    for name, module in model._modules.items():
        if len(list(module.children())) > 0:
            model._modules[name] = replace_activation(module, activation_cfg)
        if name == "activ":
            if activation_cfg["type"] == "torch_swish":
                model._modules[name] = nn.SiLU()
            else:
                model._modules[name] = build_activation_layer(activation_cfg)
    return model


def replace_norm(model, cfg):
    for name, module in model._modules.items():
        if len(list(module.children())) > 0:
            model._modules[name] = replace_norm(module, cfg)
        if name == "bn":
            model._modules[name] = build_norm_layer(cfg, num_features=module.num_features)[1]
    return model


def generate_backbones():
    logger = get_root_logger()

    for model_name, model_getter in _models.items():

        def closure(model_name, model_getter):
            def multioutput_forward(self, x):
                outputs = []
                y = x

                last_stage = max(self.out_indices)
                for i, stage in enumerate(self.features):
                    y = stage(y)
                    s = str(i) + " " + str(y.shape)
                    if i in self.out_indices:
                        outputs.append(y)
                        s += "*"
                    if self.verbose:
                        print(s)
                    if i == last_stage:
                        break

                return outputs

            def init_weights(self, pretrained=True):
                if pretrained:
                    rank, world_size = get_dist_info()
                    logger.warning(f"imgclsmob::loading weights proc rank {rank}")
                    if rank == 0:
                        # Make sure that model is fetched to the local storage.
                        logger.warning(f"imgclsmob::downloading {rank}")
                        download_model(
                            net=self, model_name=model_name, local_model_store_dir_path=self.models_cache_root
                        )
                        if world_size > 1:
                            logger.warning(f"imgclsmob::barrier {rank}")
                            distributed.barrier()
                    else:
                        # Wait for model to be in the local storage, then load it.
                        logger.warning(f"imgclsmob::barrier {rank}")
                        distributed.barrier()
                        logger.warning(f"imgclsmob::loading {rank}")
                        download_model(
                            net=self, model_name=model_name, local_model_store_dir_path=self.models_cache_root
                        )
                    logger.warning(f"imgclsmob::done {rank}")

            def train(self, mode=True):
                super(self.__class__, self).train(mode)

                for i in range(self.frozen_stages + 1):
                    m = self.features[i]
                    m.eval()
                    for param in m.parameters():
                        param.requires_grad = False

                if mode and self.norm_eval:
                    for m in self.modules():
                        # trick: eval have effect on BatchNorm only
                        if isinstance(m, _BatchNorm):
                            m.eval()

            class custom_model_getter(nn.Module):
                def __init__(
                    self,
                    *args,
                    out_indices=None,
                    frozen_stages=0,
                    norm_eval=False,
                    verbose=False,
                    activation_cfg=None,
                    norm_cfg=None,
                    **kwargs,
                ):
                    super().__init__()
                    models_cache_root = kwargs.get("root", os.path.join("~", ".torch", "models"))
                    is_pretrained = kwargs.get("pretrained", False)
                    logger.warning(
                        f"Init model {model_name}, pretrained={is_pretrained}, models cache {models_cache_root}"
                    )
                    # if 'pretrained' in kwargs and kwargs['pretrained']:
                    #     rank, _ = get_dist_info()
                    #     if rank > 0:
                    #         if 'root' not in kwargs:
                    #             kwargs['root'] = tempfile.mkdtemp()
                    #         kwargs['root'] = tempfile.mkdtemp(dir=kwargs['root'])
                    #         logger.info('Rank: {}, Setting {} as a target location of pretrained models'.format(rank, kwargs['root']))
                    model = model_getter(*args, **kwargs)
                    if activation_cfg:
                        model = replace_activation(model, activation_cfg)
                    if norm_cfg:
                        model = replace_norm(model, norm_cfg)
                    model.out_indices = out_indices
                    model.frozen_stages = frozen_stages
                    model.norm_eval = norm_eval
                    model.verbose = verbose
                    model.models_cache_root = models_cache_root
                    if hasattr(model, "features") and isinstance(model.features, nn.Sequential):
                        # Save original forward, just in case.
                        model.forward_single_output = model.forward
                        model.forward = multioutput_forward.__get__(model)
                        model.init_weights = init_weights.__get__(model)
                        model.train = train.__get__(model)

                        # model.forward = types.MethodType(multioutput_forward, model)
                        # model.init_weights = types.MethodType(init_weights, model)
                        # model.train = types.MethodType(train, model)

                        model.output = None
                        for i, _ in enumerate(model.features):
                            if i > max(out_indices):
                                model.features[i] = None
                    else:
                        raise ValueError(
                            "Failed to automatically wrap backbone network. "
                            "Object of type {} has no valid attribute called "
                            '"features".'.format(model.__class__)
                        )
                    self.__dict__.update(model.__dict__)

            custom_model_getter.__name__ = model_name
            return custom_model_getter

        BACKBONES.register_module(name=model_name, module=closure(model_name, model_getter))


generate_backbones()



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import os

import numpy as np
import pytest
from openvino.model_zoo.model_api.models import Model

import otx.algorithms.segmentation.tasks.openvino
from otx.algorithms.segmentation.configs.base import SegmentationConfig
from otx.algorithms.segmentation.tasks.openvino import (
    OpenVINOSegmentationInferencer,
    OpenVINOSegmentationTask,
)
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.metrics import Performance, ScoreMetric
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from otx.api.utils.shape_factory import ShapeFactory
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_SEG_TEMPLATE_DIR,
    generate_otx_dataset,
    generate_otx_label_schema,
    init_environment,
)


class TestOpenVINOSegmentationInferencer:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        model_template = parse_model_template(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        seg_params = SegmentationConfig(header=hyper_parameters.header)
        label_schema = generate_otx_label_schema()
        mocker.patch("otx.algorithms.segmentation.tasks.openvino.OpenvinoAdapter")
        mocker.patch.object(Model, "create_model")
        self.seg_ov_inferencer = OpenVINOSegmentationInferencer(seg_params, label_schema, "")
        self.seg_ov_inferencer.model = mocker.patch("openvino.model_zoo.model_api.models.Model", autospec=True)

        self.fake_input = np.full((5, 1), 0.1)

    @e2e_pytest_unit
    def test_pre_process(self):
        self.seg_ov_inferencer.model.preprocess.return_value = {"foo": "bar"}
        returned_value = self.seg_ov_inferencer.pre_process(self.fake_input)

        assert returned_value == {"foo": "bar"}

    @e2e_pytest_unit
    def test_post_process(self):
        fake_prediction = {"pred": self.fake_input}
        fake_metadata = {"soft_prediction": self.fake_input, "feature_vector": None}
        self.seg_ov_inferencer.model.postprocess.return_value = np.ones((5, 1))
        returned_value = self.seg_ov_inferencer.post_process(fake_prediction, fake_metadata)

        assert len(returned_value) == 3
        assert np.array_equal(returned_value[2], self.fake_input)

    @e2e_pytest_unit
    def test_predict(self, mocker):
        fake_output = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=[])
        mock_pre_process = mocker.patch.object(OpenVINOSegmentationInferencer, "pre_process", return_value=("", ""))
        mock_forward = mocker.patch.object(OpenVINOSegmentationInferencer, "forward")
        mock_post_process = mocker.patch.object(
            OpenVINOSegmentationInferencer, "post_process", return_value=fake_output
        )
        returned_value = self.seg_ov_inferencer.predict(self.fake_input)

        mock_pre_process.assert_called_once()
        mock_forward.assert_called_once()
        mock_post_process.assert_called_once()
        assert returned_value == fake_output

    @e2e_pytest_unit
    def test_forward(self):
        fake_output = {"pred": np.full((5, 1), 0.9)}
        self.seg_ov_inferencer.model.infer_sync.return_value = fake_output
        returned_value = self.seg_ov_inferencer.forward({"image": self.fake_input})

        assert returned_value == fake_output


class TestOpenVINOSegmentationTask:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, otx_model) -> None:
        model_template = parse_model_template(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        label_schema = generate_otx_label_schema()
        task_env = init_environment(hyper_parameters, model_template)
        seg_params = SegmentationConfig(header=hyper_parameters.header)
        mocker.patch("otx.algorithms.segmentation.tasks.openvino.OpenvinoAdapter")
        mocker.patch.object(Model, "create_model")
        seg_ov_inferencer = OpenVINOSegmentationInferencer(seg_params, label_schema, "")

        task_env.model = otx_model
        mocker.patch.object(OpenVINOSegmentationTask, "load_inferencer", return_value=seg_ov_inferencer)
        self.seg_ov_task = OpenVINOSegmentationTask(task_env)

    @e2e_pytest_unit
    def test_infer(self, mocker):
        self.dataset = generate_otx_dataset()
        fake_annotation = [
            Annotation(
                Polygon(points=[Point(0, 0)]),
                id=0,
                labels=[ScoredLabel(LabelEntity(name="fake", domain="SEGMENTATION"), probability=1.0)],
            )
        ]
        fake_ann_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=fake_annotation)
        fake_input = mocker.MagicMock()
        mock_predict = mocker.patch.object(
            OpenVINOSegmentationInferencer, "predict", return_value=(fake_ann_scene, None, fake_input)
        )
        mocker.patch("otx.algorithms.segmentation.tasks.openvino.get_activation_map", return_value=np.zeros((5, 1)))
        mocker.patch.object(ShapeFactory, "shape_produces_valid_crop", return_value=True)
        updated_dataset = self.seg_ov_task.infer(self.dataset)

        mock_predict.assert_called()
        for updated in updated_dataset:
            assert updated.annotation_scene.contains_any([LabelEntity(name="fake", domain="SEGMENTATION")])

    @e2e_pytest_unit
    def test_evaluate(self, mocker):
        result_set = ResultSetEntity(
            model=None,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.dice.DiceAverage", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics="mDice"
        )
        mocker.patch.object(MetricsHelper, "compute_dice_averaged_over_pixels", return_value=fake_metrics)
        self.seg_ov_task.evaluate(result_set)

        assert result_set.performance.score.value == 0.1

    @e2e_pytest_unit
    def test_deploy(self, otx_model):
        output_model = copy.deepcopy(otx_model)
        self.seg_ov_task.model.set_data("openvino.bin", b"foo")
        self.seg_ov_task.model.set_data("openvino.xml", b"bar")
        self.seg_ov_task.deploy(output_model)

        assert output_model.exportable_code is not None

    @e2e_pytest_unit
    def test_optimize(self, mocker, otx_model):
        def patch_save_model(model, dir_path, model_name):
            with open(f"{dir_path}/{model_name}.xml", "wb") as f:
                f.write(b"foo")
            with open(f"{dir_path}/{model_name}.bin", "wb") as f:
                f.write(b"bar")

        dataset = generate_otx_dataset()
        output_model = copy.deepcopy(otx_model)
        self.seg_ov_task.model.set_data("openvino.bin", b"foo")
        self.seg_ov_task.model.set_data("openvino.xml", b"bar")
        mocker.patch("otx.algorithms.segmentation.tasks.openvino.load_model", autospec=True)
        mocker.patch("otx.algorithms.segmentation.tasks.openvino.create_pipeline", autospec=True)
        mocker.patch("otx.algorithms.segmentation.tasks.openvino.save_model", new=patch_save_model)
        spy_compress = mocker.spy(otx.algorithms.segmentation.tasks.openvino, "compress_model_weights")
        self.seg_ov_task.optimize(OptimizationType.POT, dataset=dataset, output_model=output_model)

        spy_compress.assert_called_once()
        assert self.seg_ov_task.model.get_data("openvino.bin")
        assert self.seg_ov_task.model.get_data("openvino.xml")
